# coding: utf-8
from __future__ import print_function

import sys

import pytest
import xmldiff.main
from lxml import etree
from lxml.builder import E

from benker.builders.formex import FormexBuilder
from benker.cell import Cell
from benker.table import Table

TBL = E.TBL
ROW = E.ROW
P = E.P


@pytest.mark.parametrize(
    'kwargs, attrib',
    [
        ({}, {'COL': u"1"}),
        ({'width': 2}, {'COL': u"1", 'COLSPAN': u"2"}),
        ({'height': 2}, {'COL': u"1", 'ROWSPAN': u"2"}),
        ({'nature': 'body'}, {'COL': u"1", 'TYPE': 'NORMAL'}),
        ({'nature': 'header'}, {'COL': u"1", 'TYPE': 'HEADER'}),
        ({'nature': 'footer'}, {'COL': u"1", 'TYPE': '__GR.NOTES__'}),
        ({'styles': {'vertical-align': "middle"}}, {'COL': u"1"}),
        ({'styles': {'align': "center"}}, {'COL': u"1"}),
        ({'styles': {'background-color': "yellow"}}, {'COL': u"1"}),
    ],
)
def test_build_cell__body(kwargs, attrib):
    builder = FormexBuilder()

    p_elem = P(u"text")
    cell_x1_y1 = Cell([p_elem], x=1, y=1, **kwargs)
    table = Table([cell_x1_y1])
    builder._table = table

    # -- build the cell
    row_elem = ROW()
    row_y1 = next(iter(table.rows))
    builder.build_cell(row_elem, cell_x1_y1, row_y1)

    # -- check the '<CELL>' attributes
    entry_elem = row_elem[0]  # type: etree._Element
    assert entry_elem.tag == u"CELL"
    assert entry_elem.attrib == attrib
    assert entry_elem[0] == p_elem


@pytest.mark.parametrize(
    'kwargs, attrib',
    [
        ({'styles': {'vertical-align': "middle"}}, {'COL': u"1", 'valign': "middle"}),
        ({'styles': {'align': "center"}}, {'COL': u"1", 'align': "center"}),
        ({'styles': {'background-color': "yellow"}}, {'COL': u"1", 'bgcolor': "yellow"}),
    ],
)
def test_build_cell__use_cals(kwargs, attrib):
    builder = FormexBuilder(use_cals=True, cals_ns=None)

    p_elem = P(u"text")
    cell_x1_y1 = Cell([p_elem], x=1, y=1, **kwargs)
    table = Table([cell_x1_y1])
    builder._table = table

    # -- build the cell
    row_elem = ROW()
    row_y1 = next(iter(table.rows))
    builder.build_cell(row_elem, cell_x1_y1, row_y1)

    # -- check the '<CELL>' attributes
    entry_elem = row_elem[0]  # type: etree._Element
    assert entry_elem.tag == u"CELL"
    assert entry_elem.attrib == attrib
    assert entry_elem[0] == p_elem


@pytest.mark.parametrize(
    'kwargs, attrib',
    [
        ({'nature': 'body'}, {'COL': u"1", 'TYPE': 'NORMAL'}),
        ({'nature': 'header'}, {'COL': u"1"}),
        ({'nature': 'footer'}, {'COL': u"1", 'TYPE': '__GR.NOTES__'}),
    ],
)
def test_build_cell__head(kwargs, attrib):
    builder = FormexBuilder()

    p_elem = P(u"text")
    cell_x1_y1 = Cell([p_elem], x=1, y=1, **kwargs)
    table = Table([cell_x1_y1])
    builder._table = table

    # -- build the cell
    row_elem = ROW()
    row_y1 = next(iter(table.rows))
    row_y1.nature = "header"
    builder.build_cell(row_elem, cell_x1_y1, row_y1)

    # -- check the '<CELL>' attributes
    entry_elem = row_elem[0]  # type: etree._Element
    assert entry_elem.tag == u"CELL"
    assert entry_elem.attrib == attrib
    assert entry_elem[0] == p_elem


def test_build_title():
    table = Table()
    table.rows[1].insert_cell(u"Title", styles={"align": "center"})

    builder = FormexBuilder()
    tbl_elem = TBL()
    builder.build_title(tbl_elem, table.rows[0])

    # -- check the '<TITLE>' attributes
    title_elem = tbl_elem[0]  # type: etree._Element
    xml_parser = etree.XMLParser(remove_blank_text=True)
    # fmt: off
    expected = etree.XML(u"""\
    <TITLE>
      <TI>
        <P>Title</P>
      </TI>
    </TITLE>""", parser=xml_parser)
    # fmt: on

    diff_list = xmldiff.main.diff_trees(title_elem, expected)
    if diff_list:
        print(etree.tounicode(title_elem, pretty_print=True, with_tail=False), file=sys.stderr)
        assert diff_list == []


def test_build_title__empty():
    table = Table()
    table.rows[1].insert_cell(None, styles={"align": "center"})

    builder = FormexBuilder()
    tbl_elem = TBL()
    builder.build_title(tbl_elem, table.rows[0])

    # -- check the '<TITLE>' attributes
    title_elem = tbl_elem[0]  # type: etree._Element
    xml_parser = etree.XMLParser(remove_blank_text=True)
    # fmt: off
    expected = etree.XML(u"""\
    <TITLE>
      <TI>
        <IE/>
      </TI>
    </TITLE>""", parser=xml_parser)
    # fmt: on

    diff_list = xmldiff.main.diff_trees(title_elem, expected)
    if diff_list:
        print(etree.tounicode(title_elem, pretty_print=True, with_tail=False), file=sys.stderr)
        assert diff_list == []


def test_build_title__subtitle():
    table = Table()
    content = [P(u"TITLE"), P(u"Subtitle 1"), P(u"Subtitle 2")]
    table.rows[1].insert_cell(content, styles={"align": "center"})

    builder = FormexBuilder()
    tbl_elem = TBL()
    builder.build_title(tbl_elem, table.rows[0])

    # -- check the '<TITLE>' attributes
    title_elem = tbl_elem[0]  # type: etree._Element
    xml_parser = etree.XMLParser(remove_blank_text=True)
    # fmt: off
    expected = etree.XML(u"""\
    <TITLE>
      <TI>
        <P>TITLE</P>
      </TI>
      <STI>
        <P>Subtitle 1</P>
        <P>Subtitle 2</P>
      </STI>
    </TITLE>""", parser=xml_parser)
    # fmt: on

    diff_list = xmldiff.main.diff_trees(title_elem, expected)
    if diff_list:
        print(etree.tounicode(title_elem, pretty_print=True, with_tail=False), file=sys.stderr)
        assert diff_list == []


def test_build_tbl():
    # see: formex-4/samples/jo-compl-2002C_061/C_2002061EN.01000403.xml

    table = Table()
    table.rows[1].nature = "header"
    table.rows[1].insert_cell([P(u"Expert group")])
    table.rows[1].insert_cell([P(u"First name and surname of the expert")])
    table.rows[2].insert_cell([P(u"Control of infectious diseases")])
    table.rows[2].insert_cell([P(u"Michael Angelo BORG")])
    table.rows[3].insert_cell([P(u"Information society")], height=3)
    table.rows[3].insert_cell([P(u"Tony HEY")])
    table.rows[4].insert_cell([P(u"José L. ENCARNAÇÃO")])
    table.rows[5].insert_cell([P(u"Berit SVENDSEN")])
    table.rows[6].insert_cell([P(u"Controlled thermonuclear fusion")])
    table.rows[6].insert_cell([P(u"Pekka PIRILÄ")])

    builder = FormexBuilder()
    table_elem = builder.build_tbl(table)

    xml_parser = etree.XMLParser(remove_blank_text=True)

    # fmt: off
    expected = etree.XML(u"""\
    <TBL COLS="2" NO.SEQ="0001">
      <CORPUS>
        <ROW TYPE="HEADER">
          <CELL COL="1">
            <P>Expert group</P>
          </CELL>
          <CELL COL="2">
            <P>First name and surname of the expert</P>
          </CELL>
        </ROW>
        <ROW>
          <CELL COL="1">
            <P>Control of infectious diseases</P>
          </CELL>
          <CELL COL="2">
            <P>Michael Angelo BORG</P>
          </CELL>
        </ROW>
        <ROW>
          <CELL COL="1" ROWSPAN="3">
            <P>Information society</P>
          </CELL>
          <CELL COL="2">
            <P>Tony HEY</P>
          </CELL>
        </ROW>
        <ROW>
          <CELL COL="2">
            <P>José L. ENCARNAÇÃO</P>
          </CELL>
        </ROW>
        <ROW>
          <CELL COL="2">
            <P>Berit SVENDSEN</P>
          </CELL>
        </ROW>
        <ROW>
          <CELL COL="1">
            <P>Controlled thermonuclear fusion</P>
          </CELL>
          <CELL COL="2">
            <P>Pekka PIRILÄ</P>
          </CELL>
        </ROW>
      </CORPUS>
    </TBL>""", parser=xml_parser)
    # fmt: on

    for elem in table_elem.xpath("//*"):
        elem.text = elem.text or None
    for elem in expected.xpath("//*"):
        elem.text = elem.text or None

    diff_list = xmldiff.main.diff_trees(table_elem, expected)
    if diff_list:
        print(etree.tounicode(table_elem, pretty_print=True, with_tail=False), file=sys.stderr)
        assert diff_list == []


def test_build_tbl__with_title():
    # see: formex-4/samples/jo-compl-2002C_280/C_2002280EN.01000101.xml

    table = Table()
    table.rows[1].insert_cell([P(u"1 euro =")], width=3, styles={"align": "center"})
    table.rows[2].nature = "header"
    table.rows[2].insert_cell([P()], styles={"x-cell-empty": "true"})
    table.rows[2].insert_cell([P(u"Currency")])
    table.rows[2].insert_cell([P(u"Exchange rate")])
    table.rows[3].insert_cell([P(u"USD")])
    table.rows[3].insert_cell([P(u"US dollar")])
    table.rows[3].insert_cell([P(u"1,0029")])
    table.rows[4].insert_cell([P(u"JPY")])
    table.rows[4].insert_cell([P(u"Japanese yen")])
    table.rows[4].insert_cell([P(u"121,05")])

    builder = FormexBuilder(detect_titles=True)
    table_elem = builder.build_tbl(table)

    xml_parser = etree.XMLParser(remove_blank_text=True)

    # fmt: off
    expected = etree.XML(u"""\
    <TBL COLS="3" NO.SEQ="0001">
      <TITLE>
        <TI>
          <P>1 euro =</P>
        </TI>
        <STI/>
      </TITLE>
      <CORPUS>
        <ROW TYPE="HEADER">
          <CELL COL="1">
            <IE/>
          </CELL>
          <CELL COL="2">
            <P>Currency</P>
          </CELL>
          <CELL COL="3">
            <P>Exchange rate</P>
          </CELL>
        </ROW>
        <ROW>
          <CELL COL="1">
            <P>USD</P>
          </CELL>
          <CELL COL="2">
            <P>US dollar</P>
          </CELL>
          <CELL COL="3">
            <P>1,0029</P>
          </CELL>
        </ROW>
        <ROW>
          <CELL COL="1">
            <P>JPY</P>
          </CELL>
          <CELL COL="2">
            <P>Japanese yen</P>
          </CELL>
          <CELL COL="3">
            <P>121,05</P>
          </CELL>
        </ROW>
      </CORPUS>
    </TBL>""", parser=xml_parser)
    # fmt: on

    for elem in table_elem.xpath("//*"):
        elem.text = elem.text or None
    for elem in expected.xpath("//*"):
        elem.text = elem.text or None

    diff_list = xmldiff.main.diff_trees(table_elem, expected)
    if diff_list:
        print(etree.tounicode(table_elem, pretty_print=True, with_tail=False), file=sys.stderr)
        assert diff_list == []


@pytest.mark.parametrize(
    'orient, size, expected',
    [
        ('portrait', (595, 841), {'NO.SEQ': '0001', 'COLS': '1'}),
        ('landscape', (595, 841), {'NO.SEQ': '0001', 'COLS': '1', 'PAGE.SIZE': 'SINGLE.LANDSCAPE'}),
        ('portrait', (841, 595), {'NO.SEQ': '0001', 'COLS': '1'}),
        ('landscape', (841, 595), {'NO.SEQ': '0001', 'COLS': '1', 'PAGE.SIZE': 'SINGLE.LANDSCAPE'}),
        ('portrait', (1190, 841), {'NO.SEQ': '0001', 'COLS': '1', 'PAGE.SIZE': 'DOUBLE.PORTRAIT'}),
        ('landscape', (1190, 841), {'NO.SEQ': '0001', 'COLS': '1', 'PAGE.SIZE': 'DOUBLE.LANDSCAPE'}),
        ('portrait', (841, 1190), {'NO.SEQ': '0001', 'COLS': '1', 'PAGE.SIZE': 'DOUBLE.PORTRAIT'}),
        ('landscape', (841, 1190), {'NO.SEQ': '0001', 'COLS': '1', 'PAGE.SIZE': 'DOUBLE.LANDSCAPE'}),
    ],
)
def test_build_tbl__orient(orient, size, expected):
    builder = FormexBuilder()
    table = Table(styles={'x-sect-orient': orient, 'x-sect-size': size})
    table.rows[1].insert_cell(u"text")
    table_elem = builder.build_tbl(table)
    assert table_elem.attrib == expected


def test_build_tbl__no_seq():
    builder = FormexBuilder()
    table1 = Table()
    table1.rows[1].insert_cell(u"text1")
    table1_elem = builder.build_tbl(table1)
    table2 = Table()
    table2.rows[1].insert_cell(u"text2")
    table2_elem = builder.build_tbl(table2)
    assert table1_elem.attrib['NO.SEQ'] == u"0001"
    assert table2_elem.attrib['NO.SEQ'] == u"0002"


def test_build_tbl__empty_cell():
    builder = FormexBuilder()
    table1 = Table()
    table1.rows[1].insert_cell(u"")
    table1_elem = builder.build_tbl(table1)
    cell_elem = table1_elem.xpath('//CELL')[0]
    assert len(cell_elem) == 1
    assert cell_elem[0].tag == 'IE'


def test_build_tbl__use_cals():
    # see: formex-4/samples/jo-compl-2002C_061/C_2002061EN.01000403.xml

    table = Table(
        styles={
            "border-top": "solid",
            "border-bottom": "solid",
            "x-sect-orient": "landscape",
            "x-sect-cols": "1",
            "background-color": "blue",
            "width": "180",
        }
    )
    table.rows[1].nature = "header"
    table.rows[1].insert_cell([P(u"Expert group")], styles={"align": "center"})
    table.rows[1].insert_cell([P(u"First name and surname of the expert")], styles={"align": "center"})
    table.rows[2].insert_cell([P(u"Control of infectious diseases")])
    table.rows[2].insert_cell([P(u"Michael Angelo BORG")])

    builder = FormexBuilder(use_cals=True, cals_ns=None)
    table_elem = builder.build_tbl(table)

    xml_parser = etree.XMLParser(remove_blank_text=True)

    # fmt: off
    expected = etree.XML(u"""\
    <TBL NO.SEQ="0001" COLS="2" PAGE.SIZE="SINGLE.LANDSCAPE">
      <CORPUS frame="topbot" colsep="0" rowsep="0" orient="land" pgwide="1" bgcolor="blue" width="180.00mm">
        <colspec colname="c1" colnum="1"/>
        <colspec colname="c2" colnum="2"/>
        <ROW TYPE="HEADER">
          <CELL COL="1" align="center">
            <P>Expert group</P>
          </CELL>
          <CELL COL="2" align="center">
            <P>First name and surname of the expert</P>
          </CELL>
        </ROW>
        <ROW>
          <CELL COL="1">
            <P>Control of infectious diseases</P>
          </CELL>
          <CELL COL="2">
            <P>Michael Angelo BORG</P>
          </CELL>
        </ROW>
      </CORPUS>
    </TBL>""", parser=xml_parser)
    # fmt: on

    for elem in table_elem.xpath("//*"):
        elem.text = elem.text or None
    for elem in expected.xpath("//*"):
        elem.text = elem.text or None

    diff_list = xmldiff.main.diff_trees(table_elem, expected)
    if diff_list:
        print(etree.tounicode(table_elem, pretty_print=True, with_tail=False), file=sys.stderr)
        assert diff_list == []



# coding: utf-8
"""
lxml - QName
============

Python alternative to :class:`lxml.etree.QName` for lxml < 4

.. versionadded:: 0.5.0
"""
from lxml import etree

if etree.LXML_VERSION >= (4,):
    QName = etree.QName

else:
    class _QName(etree.QName):
        __doc__ = etree.QName.__doc__

        def __init__(self, text_or_uri_or_element, tag=None):
            if text_or_uri_or_element is None:
                super(_QName, self).__init__(tag)
            else:
                super(_QName, self).__init__(text_or_uri_or_element, tag=tag)

    QName = _QName



# coding: utf-8
from __future__ import print_function

import shutil

import py.path  # type hints
import pytest
import xmldiff.main
from lxml import etree

from benker.converters.formex2cals import convert_formex2cals

from tests.resources import RESOURCES_DIR


@pytest.mark.parametrize(
    'input_name, expected_name, embed_gr_notes',
    [
        ("formex/tbl_small_table.xml", "formex2cals/tbl_small_table.xml", False),
        ("formex/tbl_sample.xml", "formex2cals/tbl_sample.xml", False),
        ("formex/tbl_sample_cals.xml", "formex2cals/tbl_sample_cals.xml", False),
        ("formex/tbl_sample.xml", "formex2cals/tbl_sample.embedded.xml", True),
        ("formex/tbl_sample_cals.xml", "formex2cals/tbl_sample_cals.embedded.xml", True),
        ("formex/tbl_col_type_header.xml", "formex2cals/tbl_col_type_header.xml", False),
    ],
)
def test_convert_formex2cals(input_name, expected_name, embed_gr_notes, tmpdir):
    # type: (str, str, bool, py.path.local) -> None
    src_xml = RESOURCES_DIR.join(input_name)  # type: py.path.local
    dst_xml = tmpdir.join(src_xml.basename)
    convert_formex2cals(
        str(src_xml),
        str(dst_xml),
        width_unit='mm',
        cals_prefix="cals",
        cals_ns="https://lib.benker.com/schemas/cals.xsd",
        embed_gr_notes=embed_gr_notes,
    )

    # - Compare with expected
    xml_parser = etree.XMLParser(remove_blank_text=True)
    expected_xml = RESOURCES_DIR.join(expected_name)  # type: py.path.local
    if expected_xml.exists():
        expected_tree = etree.parse(str(expected_xml), parser=xml_parser)
        NS = {"cals": "https://lib.benker.com/schemas/cals.xsd"}
        expected_elements = expected_tree.xpath("//cals:table", namespaces=NS)
        dst_tree = etree.parse(str(dst_xml), parser=xml_parser)
        dst_elements = dst_tree.xpath("//cals:table", namespaces=NS)
        assert len(expected_elements) == len(dst_elements)

        for dst_elem, expected_elem in zip(dst_elements, expected_elements):
            diff_list = xmldiff.main.diff_trees(dst_elem, expected_elem)
            assert not diff_list
    else:
        # missing resource: create it
        shutil.copy(str(dst_xml), str(expected_xml))
        print("You should check and commit the file: '{}'".format(expected_xml))



# coding: utf-8
"""
Drawing
=======

Utility functions used to draw a :class:`~benker.grid.Grid`.
"""
from benker.box import Box

#: Default tiles used to draw a :class:`~benker.grid.Grid`.
#:
#: Keys are tuples (*left*, *top*, *right*, *bottom*) : which represent the
#: presence (if ``True``) or absence (if ``False``) : of the border.
#: Values are the string representation of the tiles,
#: "XXXXXXXXX" will be replaced by the cell content.
TILES = {
    (True, True, True, True): """\
+-----------+
| XXXXXXXXX |
+-----------+
""",
    (True, True, True, False): """\
+-----------+
| XXXXXXXXX |
""",
    (True, True, False, True): """\
+-----------
| XXXXXXXXX 
+-----------
""",
    (True, True, False, False): """\
+-----------
| XXXXXXXXX 
""",

    (True, False, True, True): """\
|           |
| XXXXXXXXX |
+-----------+
""",
    (True, False, True, False): """\
|           |
| XXXXXXXXX |
""",
    (True, False, False, True): """\
|           
| XXXXXXXXX 
+-----------
""",
    (True, False, False, False): """\
|           
| XXXXXXXXX 
""",

    (False, True, True, True): """\
------------+
  XXXXXXXXX |
------------+
""",
    (False, True, True, False): """\
------------+
  XXXXXXXXX |
""",
    (False, True, False, True): """\
------------
  XXXXXXXXX 
------------
""",
    (False, True, False, False): """\
------------
  XXXXXXXXX 
""",

    (False, False, True, True): """\
            |
  XXXXXXXXX |
------------+
""",
    (False, False, True, False): """\
            |
  XXXXXXXXX |
""",
    (False, False, False, True): """\
            
  XXXXXXXXX 
------------
""",
    (False, False, False, False): """\
            
  XXXXXXXXX 
""",
}


def iter_tiles(grid, tiles=None):
    tiles = tiles or TILES
    bb = grid.bounding_box
    for row_idx in range(bb.min.y, bb.max.y + 1):
        row = []
        for col_idx in range(bb.min.x, bb.max.x + 1):
            coord = col_idx, row_idx
            if coord in grid:
                cell = grid[coord]
                box = cell.box
                text = str(cell)
            else:
                box = Box(col_idx, row_idx)
                text = ""
            left = box.min.x == col_idx
            top = box.min.y == row_idx
            right = bb.max.x == col_idx
            bottom = bb.max.y == row_idx
            tile = tiles[(left, top, right, bottom)]
            if (box.min.x + box.max.x) // 2 == col_idx and (box.min.y + box.max.y) // 2 == row_idx:
                title = "{0:^9}".format(str(text))[:9]
                tile = tile.replace('XXXXXXXXX', title)
            else:
                tile = tile.replace('XXXXXXXXX', ' ' * 9)
            row.append(tile)
        yield row


def iter_lines(grid, tiles=None):
    for row in iter_tiles(grid, tiles):
        tiles = [list(filter(None, tile.splitlines())) for tile in row]
        size = len(tiles[0])
        for index in range(size):
            yield "".join(tile[index] for tile in tiles)


def draw(grid, tiles=None):
    """
    Draw a grid using a collection of tiles.

    :type  grid: benker.grid.Grid
    :param grid: Grid to draw.

    :param tiles:
        Collection of tiles, use :data:`~benker.drawing.TILES` if not provided.

    :return: String representation of the grid.
    """
    if grid:
        return "\n".join(iter_lines(grid, tiles))
    return ""



# coding: utf-8
"""
CALS Parser Implementation
==========================

This module can parse the tables (``table`` elements) of a CALS file.

Specifications and examples:

- The CALS DTD is available online in the OASIS web site:
  `CALS Table Model Document Type Definition <https://www.oasis-open.org/specs/a502.htm>`_.

- An example of CALS table is available in Wikipedia:
  `CALS Table Model <https://en.wikipedia.org/wiki/CALS_Table_Model>`_

The main elements of a CALS table are:

*   ``table``: a table can contains one or several ``tgroup``.

    *   ``titles``: table titles (*not supported by the CALS parser*)

    *   ``tgroup``: a portion of table

        *   ``colspec``: column specifications

        *   ``spanspec``: spanning specifications (*not supported by the CALS parser*)

        *   ``thead``: table header

            *   ``colspec``: header column specifications (*not supported by the CALS parser*)

            *   ``row``: table row (see ``tbody``)

        *   ``tfoot``: table footer

            *   ``colspec``: footer column specifications (*not supported by the CALS parser*)

            *   ``row``: table row (see ``tbody``)

        *   ``tbody``: table body

            *   ``row``: table row

                *   ``entry``: table entry which contains paragraphs

                *   ``entrytbl``: table entry which contains a table (*not supported by the CALS parser*)

An example of CALS table is available in Wikipedia:
`CALS Table Model <https://en.wikipedia.org/wiki/CALS_Table_Model>`_

.. versionadded:: 0.5.0
"""
import re

from lxml import etree

from benker.box import Box
from benker.common.lxml_iterwalk import iterwalk
from benker.common.lxml_qname import QName
from benker.parsers.base_parser import BaseParser
from benker.parsers.cals.frame_styles import BORDER_NONE
from benker.parsers.cals.frame_styles import BORDER_SOLID
from benker.parsers.cals.frame_styles import get_frame_styles
from benker.table import Table
from benker.units import convert_value
from benker.units import parse_width

# noinspection PyProtectedMember
#: Element Type
ElementType = etree._Element


class CalsParser(BaseParser):
    """
    CALS tables parser
    """

    def __init__(self, builder, cals_ns=None, width_unit="mm", **options):
        """
        Construct a parser

        :type  builder: benker.builders.base_builder.BaseBuilder
        :param builder:
            Builder used by this parser to instantiate :class:`~benker.table.Table` objects.

        :param str cals_ns:
            Namespace to use for CALS elements and attributes.
            Set ``None`` (or "") if you don't use namespace in your XML.

        :param str width_unit:
            Unit to use for table/column widths.
            Possible values are: 'cm', 'dm', 'ft', 'in', 'm', 'mm', 'pc', 'pt', 'px'.

        :keyword str options: Extra conversion options.
            See :meth:`~benker.converters.base_converter.BaseConverter.convert_file`
            to have a list of all possible options.

        .. versionchanged:: 0.5.1
           Add the options *width_unit*.
        """
        self.cals_ns = cals_ns
        self.width_unit = width_unit
        super(CalsParser, self).__init__(builder, **options)

    def get_cals_qname(self, name):
        return QName(self.cals_ns, name)

    def transform_tables(self, tree):
        if self.cals_ns:
            nodes = tree.xpath("//cals:table", namespaces={"cals": self.cals_ns})
        else:
            nodes = tree.xpath("//table")
        for node in nodes:
            table = self.parse_table(node)
            table_elem = self.builder.generate_table_tree(table)
            parent = node.getparent()
            index = parent.index(node)
            parent.insert(index, table_elem)
            table_elem.tail = node.tail
            parent.remove(node)

    # noinspection PyPep8Naming
    def parse_table(self, cals_table):
        """
        Convert a ``<table>`` CALS element into table object.

        :type  cals_table: ElementType
        :param cals_table: CALS element.

        :rtype: benker.table.Table
        :return: Table.
        """
        state = self._state
        state.reset()

        # -- CALS elements
        cals = self.get_cals_qname

        table = cals("table").text
        # titles = cals("titles").text  # not supported
        tgroup = cals("tgroup").text
        colspec = cals("colspec").text
        thead = cals("thead").text
        tfoot = cals("tfoot").text
        tbody = cals("tbody").text
        row = cals("row").text
        # entrytbl = cals("entrytbl").text  # not supported
        entry = cals("entry").text

        elements = {table, tgroup, colspec, thead, tfoot, tbody, row, entry}
        context = iterwalk(cals_table, events=("start", "end"), tag=elements)

        depth = 0
        for action, elem in context:
            elem_tag = elem.tag
            if elem_tag == table:
                if action == "start":
                    depth += 1
                else:
                    depth -= 1
            if depth > 1:
                # .. note:: context.skip_subtree() is not available for all version of lxml
                # This <TBL> element is inside the table.
                # It will be handled separately in another call to transform_tables()
                continue
            if action == "start":
                # tags sorted by frequency:
                if elem_tag == entry:
                    state.next_col()
                    self.parse_cals_entry(elem)

                elif elem_tag == row:
                    state.next_row()
                    self.parse_cals_row(elem)

                elif elem_tag in {tbody, tfoot, thead}:
                    # everything is done in parse_fmx_row()
                    pass

                elif elem_tag == colspec:
                    state.next_col()
                    self.parse_cals_colspec(elem)

                elif elem_tag == tgroup:
                    self.parse_cals_tgroup(elem)

                elif elem_tag == table:
                    self.parse_cals_table(elem)

                else:
                    raise NotImplementedError(elem_tag)
            else:
                if elem_tag in {row}:
                    bounding_box = Box(1, state.row_pos, len(state.table.cols), state.row_pos)
                    state.table.fill_missing(bounding_box, None, nature=state.row.nature)
                elif elem_tag == table:
                    state.table.fill_missing(state.table.bounding_box, None)

        return state.table

    def setup_table(self, styles=None, nature=None):
        table = Table(styles=styles, nature=nature)
        self._state.table = table
        return self._state

    def parse_cals_table(self, cals_table):
        """
        Parse a CALS ``table`` element.

        :type  cals_table: ElementType
        :param cals_table: CALS table Element.

        :return: State of the parser (for debug purpose).

        .. versionchanged:: 0.5.1
           Add support for the ``@cals:width`` attribute (table width).
        """
        cals = self.get_cals_qname
        styles = {}
        nature = None

        # -- attribute @cals:frame
        frame = cals_table.attrib.get(cals("frame"))
        styles.update(get_frame_styles(frame))

        # -- attribute @cals:colsep
        colsep = cals_table.attrib.get(cals("colsep"))
        colsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if colsep in colsep_map:
            styles["x-cell-border-right"] = colsep_map[colsep]

        # -- attribute @cals:rowsep
        rowsep = cals_table.attrib.get(cals("rowsep"))
        rowsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if rowsep in rowsep_map:
            styles["x-cell-border-bottom"] = rowsep_map[rowsep]

        # -- attribute @cals:orient
        orient = cals_table.attrib.get(cals("orient"))
        orient_map = {"land": "landscape", "port": "portrait"}
        if orient in orient_map:
            styles["x-sect-orient"] = orient_map[orient]

        # -- attribute @cals:pgwide
        pgwide = cals_table.attrib.get(cals("pgwide"))
        pgwide_map = {"0": "2", "1": "1"}
        if pgwide in pgwide_map:
            styles["x-sect-cols"] = pgwide_map[pgwide]

        # -- attribute @cals:bgcolor
        bgcolor = cals_table.attrib.get(cals("bgcolor"))
        if bgcolor:
            styles["background-color"] = bgcolor

        # -- attribute @cals:tabstyle
        tabstyle = cals_table.attrib.get(cals("tabstyle"))
        if tabstyle:
            nature = tabstyle

        # -- attribute @cals:tabstyle
        width = cals_table.attrib.get(cals("width"))
        if width:
            width, unit = parse_width(width)
            value = convert_value(width, unit, self.width_unit)
            styles["width"] = u"{value:0.2f}{unit}".format(value=value, unit=self.width_unit)

        return self.setup_table(styles, nature)

    def parse_cals_tgroup(self, cals_tgroup):
        cals = self.get_cals_qname
        styles = {}
        nature = self._state.table.nature

        # -- attribute @cals:cols => ignored (*table.cols*)

        # -- attribute @cals:colsep
        colsep = cals_tgroup.attrib.get(cals("colsep"))
        colsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if colsep in colsep_map:
            styles["x-cell-border-right"] = colsep_map[colsep]

        # -- attribute @cals:rowsep
        rowsep = cals_tgroup.attrib.get(cals("rowsep"))
        rowsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if rowsep in rowsep_map:
            styles["x-cell-border-bottom"] = rowsep_map[rowsep]

        # -- attribute @cals:tgroupstyle
        tgroupstyle = cals_tgroup.attrib.get(cals("tgroupstyle"))
        if tgroupstyle:
            nature = self._state.table.nature
            if nature:
                parts = nature.split(" ")
                nature = " ".join(parts[:-1] + [tgroupstyle])
            else:
                nature = tgroupstyle

        # -- Override the table defaults
        state = self._state
        table = state.table
        table.styles.update(styles)
        table.nature = nature

        return state  # mainly for unit test

    def parse_cals_row(self, cals_row):
        """
        Parse a ``row`` element which contains ``entry`` elements.

        This element may be in a ``BLK```

        :type  cals_row: ElementType
        :param cals_row: table row

        .. versionchanged:: 0.5.1
           The "vertical-align" style is built from the ``@cals:valign`` attribute.
        """
        cals = self.get_cals_qname
        styles = {}
        nature = None  # overridden below if parent's element exists

        cals_parent = cals_row.getparent()  # type:  ElementType
        if cals_parent is not None:
            # -- nature of the row
            tag_map = {"thead": "header", "tfoot": "footer", "tbody": "body"}
            localname = QName(cals_parent.tag).localname
            nature = tag_map[localname]

            # -- attribute @cals:valign
            valign = cals_parent.attrib.get(cals("valign"))
            valign_map = {'top': 'top', 'middle': 'middle', 'bottom': 'bottom'}
            if valign in valign_map:
                styles["vertical-align"] = valign_map[valign]

        # -- attribute @cals:valign
        valign = cals_row.attrib.get(cals("valign"))
        valign_map = {'top': 'top', 'middle': 'middle', 'bottom': 'bottom'}
        if valign in valign_map:
            # overrides parent's value
            styles["vertical-align"] = valign_map[valign]

        # -- attribute @cals:rowsep
        rowsep = cals_row.attrib.get(cals("rowsep"))
        rowsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if rowsep in rowsep_map:
            styles["border-bottom"] = rowsep_map[rowsep]

        # -- attribute @cals:bgcolor
        bgcolor = cals_row.attrib.get(cals("bgcolor"))
        if bgcolor:
            styles["background-color"] = bgcolor

        # -- attribute @cals:rowstyle (extension)
        rowstyle = cals_row.attrib.get(cals("rowstyle"))
        if rowstyle:
            # overrides the previously calculated @cals:rowstyle attribute
            styles["rowstyle"] = rowstyle

        # -- Create a row
        state = self._state
        state.row = state.table.rows[state.row_pos]
        state.row.nature = nature
        state.row.styles = styles

        return state  # mainly for unit test

    def parse_cals_entry(self, cals_entry):
        """
        Parse a ``entry`` element.

        :type  cals_entry: ElementType
        :param cals_entry: table entry

        .. versionchanged:: 0.5.1
           The "vertical-align" style is built from the ``@cals:valign`` attribute.

        .. versionchanged:: 0.5.2
           Add support for the ``@cals:cellstyle`` attribute (extension).
           This attribute is required for two-way conversion of Formex tables to CALS and vice versa.
           If the ``CELL/@TYPE`` and the ``ROW/@TYPE`` are different, we add a specific "cellstyle" style.
           This style will keep the ``CELL/@TYPE`` value.

        .. versionchanged:: 0.5.3
           Improved empty cells detection for Formex4 conversion (``<IE/>`` tag management).
        """
        cals = self.get_cals_qname
        styles = {}

        # -- attribute @cals:colsep
        colsep = cals_entry.attrib.get(cals("colsep"))
        colsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if colsep in colsep_map:
            styles["border-right"] = colsep_map[colsep]

        # -- attribute @cals:rowsep
        rowsep = cals_entry.attrib.get(cals("rowsep"))
        rowsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if rowsep in rowsep_map:
            styles["border-bottom"] = rowsep_map[rowsep]

        # -- attribute @cals:bgcolor
        bgcolor = cals_entry.attrib.get(cals("bgcolor"))
        if bgcolor:
            styles["background-color"] = bgcolor

        # -- attributes @cals:namest and @cals:nameend
        name_start = cals_entry.attrib.get(cals("namest"), str(self._state.col_pos))
        name_end = cals_entry.attrib.get(cals("nameend"), str(self._state.col_pos))
        col_start = int(re.findall(r"\d+", name_start)[0])
        col_end = int(re.findall(r"\d+", name_end)[0])
        width = col_end - col_start + 1

        # -- attribute @cals:morerows
        morerows = cals_entry.attrib.get(cals("morerows"), "0")
        height = int(morerows) + 1

        # -- attribute @cals:valign
        valign = cals_entry.attrib.get(cals("valign"))
        valign_map = {'top': 'top', 'middle': 'middle', 'bottom': 'bottom'}
        if valign in valign_map:
            # overrides parent's value
            styles["vertical-align"] = valign_map[valign]

        # -- attribute @cals:rowsep
        align = cals_entry.attrib.get(cals("align"))
        align_map = {"left": "left", "right": "right", "center": "center", "justify": "justify", "char": "left"}
        if align in align_map:
            styles["align"] = align_map[align]

        # -- attribute @cals:bgcolor
        cellstyle = cals_entry.attrib.get(cals("cellstyle"))
        if cellstyle:
            styles["cellstyle"] = cellstyle

        # todo: calculate the ``@rotate`` attribute.

        # -- Create a entry
        text = [cals_entry.text] if cals_entry.text else []
        content = text + cals_entry.getchildren()

        # The detection of empty cells is used when converting
        # to the Formex4 format in order to insert an empty tag ``<IE/>``.
        # see: https://github.com/laurent-laporte-pro/benker/issues/13
        #
        # When parsing a CALS table, we will consider a cell to be empty when it contains no text
        # or child tag. As we do not know the elements contained in an ``<entry>`` element,
        # we cannot know precisely if it is an empty content or not (eg. empty paragraph or BR).
        if not content:
            styles["x-cell-empty"] = "true"

        nature = self._state.row.nature
        self._state.row.insert_cell(content, width=width, height=height, styles=styles, nature=nature)

        return self._state  # mainly for unit test

    def parse_cals_colspec(self, cals_colspec):
        """
        Parse a CALS-like ``colspec`` element.

        For instance:

        .. code-block:: xml

           <colspec
             colname="c1"
             colnum="1"
             colsep="1"
             rowsep="1"
             colwidth="30mm"
             align="center"/>

        :type  cals_colspec: ElementType
        :param cals_colspec: CALS-like ``colspec`` element.
        """
        cals_parent = cals_colspec.getparent()
        if cals_parent is not None:  # pragma: no cover
            localname = QName(cals_parent).localname
            if localname not in {"table", "tgroup"}:
                raise NotImplementedError("colspec in {} not supported".format(localname))

        cals = self.get_cals_qname
        styles = {}

        # -- attribute @cals:colname is ignored
        # -- attribute @cals:char is ignored
        # -- attribute @cals:charoff is ignored

        # -- attribute @cals:colnum
        colnum = cals_colspec.attrib.get(cals("colnum"))
        colnum = int(colnum) if colnum else self._state.col_pos

        # -- attribute @cals:colsep
        colsep = cals_colspec.attrib.get(cals("colsep"))
        colsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if colsep in colsep_map:
            styles["border-right"] = colsep_map[colsep]

        # -- attribute @cals:rowsep
        rowsep = cals_colspec.attrib.get(cals("rowsep"))
        rowsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if rowsep in rowsep_map:
            styles["border-bottom"] = rowsep_map[rowsep]

        # -- attribute @cals:rowsep
        colwidth = cals_colspec.attrib.get(cals("colwidth"))
        if colwidth:
            styles["width"] = colwidth

        # -- attribute @cals:rowsep
        align = cals_colspec.attrib.get(cals("align"))
        align_map = {"left": "left", "right": "right", "center": "center", "justify": "justify", "char": "left"}
        if align in align_map:
            styles["align"] = align_map[align]

        state = self._state
        state.col = state.table.cols[colnum]
        state.col.styles.update(styles)

        return state



# coding: utf-8
"""
Base Builder
============

Base class of Builders.
"""
import numbers

from lxml import etree

try:
    # noinspection PyCompatibility
    from collections.abc import Sequence
except ImportError:
    # noinspection PyCompatibility
    from collections import Sequence


text_type = type(u"")
binary_type = type(b"")
string_types = text_type, binary_type

# noinspection PyProtectedMember
#: Element Type
ElementType = etree._Element


class BaseBuilder(object):
    """
    Base class of Builders.
    """

    def __init__(self, **options):
        """
        Construct a builder.

        :param options: building options.
        """
        self.options = options

    def generate_table_tree(self, table):
        """
        Build the XML table from the Table instance.

        :type  table: benker.table.Table
        :param table: Table

        :return: Table tree
        """
        raise NotImplementedError

    def finalize_tree(self, tree):
        """
        Give the opportunity to finalize the resulting tree structure.

        :param tree: The resulting tree.

        .. versionadded:: 0.4.0
        """

    @staticmethod
    def append_cell_elements(cell_elem, elements):
        """
        Append XML elements, PIs or texts to a cell element.

        :type  cell_elem: ElementType
        :param cell_elem: Cell element

        :param elements: list of child elements to append

        .. versionadded:: 0.5.1
        """
        if elements is None:
            cell_elem.text = elements
        elif isinstance(elements, text_type):
            cell_elem.text = elements
        elif isinstance(elements, binary_type):
            cell_elem.text = elements.decode('utf-8')  # PY2
        elif isinstance(elements, numbers.Number):
            cell_elem.text = text_type(elements)
        elif isinstance(elements, ElementType):
            cell_elem.append(elements)
        elif isinstance(elements, Sequence):
            for node in elements:
                if isinstance(node, ElementType):
                    cell_elem.append(node)
                elif len(cell_elem) == 0:
                    text = cell_elem.text or ""
                    cell_elem.text = text + node
                else:
                    last_elem = cell_elem[-1]
                    tail = last_elem.tail or ""
                    last_elem.tail = tail + node
        else:  # pragma: no cover
            raise TypeError(repr(elements))



# coding: utf-8
"""
Table
=====

Generic table structure which simplify the conversion from docx table format to CALS or HTML tables.
"""
import re

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

from benker.cell import Cell
from benker.coord import Coord
from benker.grid import Grid
from benker.styled import Styled


# noinspection PyUnresolvedReferences
class TableViewList(object):
    """
    This class defined a (simplified) list of views.

    Short demonstration:

    .. doctest:: table_demo

        >>> from benker.cell import Cell
        >>> from benker.table import Table
        >>> from benker.table import ColView
        >>> from benker.table import RowView
        >>> from benker.table import TableViewList

        >>> red = Cell('red', x=1, y=1, height=2)
        >>> pink = Cell('pink', x=2, y=1, width=2)
        >>> blue = Cell('blue', x=2, y=2)
        >>> table = Table([red, pink, blue])
        >>> print(table)
        +-----------+-----------------------+
        |    red    |   pink                |
        |           +-----------+-----------+
        |           |   blue    |           |
        +-----------+-----------+-----------+

        >>> cols = TableViewList(table, ColView)
        >>> len(cols)
        3
        >>> rows = TableViewList(table, RowView)
        >>> len(rows)
        2

        >>> for pos, col in enumerate(cols, 1):
        ...     print("col #{pos}: {col}".format(pos=pos, col=str(col)))
        col #1: [red]
        col #2: [pink, blue]
        col #3: []

        >>> cols[3].insert_cell("yellow")
        >>> print(table)
        +-----------+-----------------------+
        |    red    |   pink                |
        |           +-----------+-----------+
        |           |   blue    |  yellow   |
        +-----------+-----------+-----------+
    """

    def __init__(self, table, view_cls):
        """
        Construct a list of view attached to the given *table*.

        :type  table: benker.table.Table
        :param table: Reference to the table.

        :type  view_cls: type
        :param view_cls:
            A view class, either :class:`~benker.table.RowView` or :class:`~benker.table.ColView`.
        """
        self.table = table
        self.view_cls = view_cls
        self.views = []
        self.refresh_all()

    def adopt_cell(self, cell):
        """
        Adopt a new cell in the views.

        :type  cell: benker.cell.Cell
        :param cell: New cell to adopt
        """
        views = self.views
        box = cell.box
        if self.view_cls is RowView:
            self._fit_to_size(box.max.y, truncate=False)
            for pos in range(box.min.y, box.max.y + 1):
                views[pos - 1].adopt_cell(cell)
        elif self.view_cls is ColView:
            self._fit_to_size(box.max.x, truncate=False)
            for pos in range(box.min.x, box.max.x + 1):
                views[pos - 1].adopt_cell(cell)
        else:
            raise TypeError(repr(self.view_cls))

    def refresh_all(self):
        """
        Cleanup and refresh all the views, taking into account the cells
        which are in the table grid.
        """
        views = self.views
        bounding_box = self.table.bounding_box
        if bounding_box is None:
            views[:] = []
            return

        if self.view_cls is RowView:
            self._fit_to_size(bounding_box.max.y)
        elif self.view_cls is ColView:
            self._fit_to_size(bounding_box.max.x)
        else:
            raise TypeError(repr(self.view_cls))

        for view in views:
            view.owned_cells[:] = []
            view.caught_cells[:] = []

        for cell in self.table:
            self.adopt_cell(cell)

    def _fit_to_size(self, size, truncate=True):
        views = self.views
        if truncate:
            del views[size:]
        for index in range(len(views), size):
            views.append(self.view_cls(self.table, index + 1))

    def __getitem__(self, pos):
        self._fit_to_size(pos, truncate=False)
        return self.views[pos - 1]

    def __len__(self):
        return len(self.views)

    def __iter__(self):
        return iter(self.views)


def _camel_to_snake(camel_name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', camel_name)
    snake_name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
    return snake_name


class ViewsProperty(object):
    """
    Descriptor used to define the rows/cols properties
    in the class :class:`~benker.table.Table`.
    """

    def __init__(self, view_cls):
        """
        Construct a "Views" property which handles a list of items which type is *view_cls*.

        :type  view_cls: type
        :param view_cls:
            A view class, either :class:`~benker.table.RowView` or :class:`~benker.table.ColView`.
        """
        self.view_cls = view_cls
        # CamelCase to snake_case:
        cls_module = self.view_cls.__module__
        cls_name = self.view_cls.__name__
        self.attr_name = "_{}".format(_camel_to_snake(cls_name))
        self.__doc__ = "List of views of type :class:`~{0}.{1}`".format(cls_module, cls_name)

    def __get__(self, table, owner):
        if table is None:
            return self
        if not hasattr(table, self.attr_name):
            setattr(table, self.attr_name, TableViewList(table, self.view_cls))
        return getattr(table, self.attr_name)


class TableView(Styled):
    """
    Base class of :class:`~benker.table.RowView` and :class:`~benker.table.ColView`
    used to create a view on the table cells for a given row or column.

    See also: :class:`~benker.table.TableViewList`
    """
    __slots__ = ('_table', '_pos', '_owned_cells', '_caught_cells')

    def __init__(self, table, pos, styles=None, nature=None):
        """
        Initialize a row/column view.

        :type  table: benker.table.Table
        :param table: Reference to the table which contains the cells to "view".

        :param int pos:
            Row or column pos (1-based) used to select cells.

        :type  styles: typing.Dict[str, str]
        :param styles:
            Dictionary of key-value pairs, where *keys* are the style names.

        :param str nature:
            Cell *nature* used to distinguish the body cells, from the header and the footer.
            The default value is ``None``, but you can use "body", "header", "footer" or whatever
            is suitable for your needs.

        .. versionchanged:: 0.4.2
           The default value of *nature* is ``None`` (instead of "body").
        """
        super(TableView, self).__init__(styles, nature)
        self._table = table
        self._pos = pos
        self._owned_cells = []
        self._caught_cells = []

    def __str__(self):
        items = ", ".join(str(cell) for cell in self._owned_cells)
        return "[{0}]".format(items)

    @property
    def table(self):
        """ Non-mutable reference to the table (instance of :class:`~benker.table.Table`). """
        return self._table

    @property
    def owned_cells(self):
        """ List of cells owned by this view. """
        return self._owned_cells

    @property
    def caught_cells(self):
        """ List of cells caught (intercepted) by this view. """
        return self._caught_cells

    def can_own(self, cell):
        """
        Check if a cell can be stored it that view.

        :type  cell: benker.cell.Cell
        :param cell: table cell
        :return: ``True`` if the cell belong to this view.
        """
        raise NotImplementedError

    def can_catch(self, cell):
        """
        Check if a cell can be caught by that view.

        :type  cell: benker.cell.Cell
        :param cell: table cell
        :return: ``True`` if the cell intercept to this view.
        """
        raise NotImplementedError

    def adopt_cell(self, cell):
        """
        Event handler called by the system when
        a cell is about to be inserted in the table.
        """
        if self.can_own(cell):
            self._owned_cells.append(cell)
        if self.can_catch(cell):
            self._caught_cells.append(cell)


class RowView(TableView):
    """
    A view on the table cells for a given row.
    """

    def can_own(self, cell):
        return cell.min.y == self._pos

    def can_catch(self, cell):
        return Coord(cell.min.x, self._pos) in cell

    @property
    def row_pos(self):
        """ Row position in the table (1-based). """
        return self._pos

    def insert_cell(self, content, styles=None, nature=None, width=1, height=1):
        """
        Insert a new cell in the row at the next free position, or at the end.

        :param content:
            User-defined cell content. It can be of any type: ``None``,
            :class:`str`, :class:`int`, :class:`float`, a container (:class:`list`),
            a XML element, etc. The same content can be shared by several cells, it's
            your own responsibility to handle the copy (or deep copy) of the *content*
            reference when needed.

        :type  styles: typing.Dict[str, str]
        :param styles:
            User-defined cell styles: a dictionary of key-value pairs.
            This values are useful to store some HTML-like styles (border-style,
            border-width, border-color, vertical-align, text-align, etc.).
            Of course, we are not tied to the HTML-like styles, you can use your own
            list of styles.

        :type nature: str
        :ivar nature: a way to distinguish the body cells, from the header and the footer.
            The default value is ``None``, but you can use "body", "header", "footer" or whatever
            is suitable for your needs.
            If set to ``None``, the cell nature is inherited from the row nature.

        :param int width:
            Width of the cell (columns spanning), default to 1.

        :param int height:
            Height of the cell (rows spanning), default to 1.

        .. versionchanged:: 0.4.2
           The *nature* of a cell is inherited from its parent's row.
        """
        caught_cells = self._caught_cells
        y = self._pos
        if caught_cells:
            boxes = tuple(cell.box for cell in caught_cells)
            bounding_box = boxes[0].union(*boxes[1:])
            for x in range(1, bounding_box.max.x + 1):
                if all((x, y) not in box for box in boxes):
                    break
            else:
                x = bounding_box.max.x + 1
        else:
            x = 1
        nature = self.nature if nature is None else nature
        cell = Cell(content, styles=styles, nature=nature,
                    x=x, y=y, width=width, height=height)
        self._table[cell.min] = cell


class ColView(TableView):
    """
    A view on the table cells for a given column.
    """

    def can_own(self, cell):
        return cell.min.x == self._pos

    def can_catch(self, cell):
        return Coord(self._pos, cell.min.y) in cell

    @property
    def col_pos(self):
        """ Column position in the table (1-based). """
        return self._pos

    def insert_cell(self, content, styles=None, nature=None, width=1, height=1):
        """
        Insert a new cell in the column at the next free position, or at the end.

        :param content:
            User-defined cell content. It can be of any type: ``None``,
            :class:`str`, :class:`int`, :class:`float`, a container (:class:`list`),
            a XML element, etc. The same content can be shared by several cells, it's
            your own responsibility to handle the copy (or deep copy) of the *content*
            reference when needed.

        :type  styles: typing.Dict[str, str]
        :param styles:
            User-defined cell styles: a dictionary of key-value pairs.
            This values are useful to store some HTML-like styles (border-style,
            border-width, border-color, vertical-align, text-align, etc.).
            Of course, we are not tied to the HTML-like styles, you can use your own
            list of styles.

        :type nature: str
        :ivar nature: a way to distinguish the body cells, from the header and the footer.
            The default value is ``None``, but you can use "body", "header", "footer" or whatever
            is suitable for your needs.
            If set to ``None``, the cell nature is inherited from the column nature.

        :param int width:
            Width of the cell (columns spanning), default to 1.

        :param int height:
            Height of the cell (rows spanning), default to 1.

        .. versionchanged:: 0.4.2
           The *nature* of a cell is inherited from its parent's column.
        """
        caught_cells = self._caught_cells
        x = self._pos
        if caught_cells:
            boxes = tuple(cell.box for cell in caught_cells)
            bounding_box = boxes[0].union(*boxes[1:])
            for y in range(1, bounding_box.max.y + 1):
                if all((x, y) not in box for box in boxes):
                    break
            else:
                y = bounding_box.max.y + 1
        else:
            y = 1
        nature = self.nature if nature is None else nature
        cell = Cell(content, styles=styles, nature=nature,
                    x=x, y=y, width=width, height=height)
        self._table[cell.min] = cell


class Table(Styled, MutableMapping):
    """
    Table data structure used to simplify conversion to CALS or HTML.

    Short demonstration:

    .. doctest:: table_demo

        >>> from benker.cell import Cell
        >>> from benker.table import Table

        >>> table = Table(styles={'frame': 'all'})

        >>> table[(1, 1)] = Cell("one")
        >>> table.rows[1].insert_cell("two")

        >>> table[(2, 1)]
        <Cell('two', styles={}, nature=None, x=2, y=1, width=1, height=1)>

        >>> table.cols[1].insert_cell("alpha")
        >>> table.cols[2].insert_cell("beta")
        >>> (1, 2) in table
        True

        >>> del table[(1, 2)]
        >>> (1, 2) in table
        False

        >>> len(table)
        3

        >>> for cell in table:
        ...     print(cell)
        one
        two
        beta

        >>> for row in table.rows:
        ...     print(row)
        [one, two]
        [beta]

        >>> table.merge((1, 2), (2, 2))
        >>> print(table)
        +-----------+-----------+
        |    one    |    two    |
        +-----------------------+
        |   beta                |
        +-----------------------+

        >>> table.expand((1, 1), width=3)
        >>> print(table)
        +-----------------------------------------------+
        |              onetwo                           |
        +-----------------------+-----------+-----------+
        |   beta                |           |           |
        +-----------------------+-----------+-----------+
    """

    rows = ViewsProperty(RowView)
    cols = ViewsProperty(ColView)

    def __init__(self, cells=None, styles=None, nature=None):
        """
        Construct a table object from a collection of cells and a dictionary of styles.

        :type  cells: typing.Iterable[benker.cell.Cell]
        :param cells: Collection of cells.

        :type  styles: typing.Dict[str, str]
        :param styles:
            Dictionary of key-value pairs, where *keys* are the style names.

        :type nature: str
        :ivar nature:
            User-defined string value.
            Table *nature* is similar to HTML ``@class`` attribute,
            you can use it do identify the styles to apply to your table.
        """
        super(Table, self).__init__(styles, nature)
        self._grid = Grid(cells)

    def __str__(self):
        return str(self._grid)

    @property
    def bounding_box(self):
        """
        Bounding box of the table (``None`` if the table is empty).

        :rtype: benker.box.Box
        :return: The bounding box.
        """
        return self._grid.bounding_box

    def _refresh_views(self, cell=None):
        """
        Refresh all the rows and column views.
        """
        if cell:
            self.rows.adopt_cell(cell)
            self.cols.adopt_cell(cell)
        else:
            self.rows.refresh_all()
            self.cols.refresh_all()

    def __contains__(self, coord):
        """
        Check if the table contains a cell (a owned cell) at the given coordinates.

        :type  coord: Coord or tuple(int, int)
        :param coord: Coordinates in the tables (1-indexed).

        :return: ``True`` if the table contains a cell at the given
            coordinates, else ``False``.
        """
        return coord in self._grid

    def __setitem__(self, coord, cell):
        """
        Insert a cell in the table at the given coordinates.

        :type  coord: Coord or tuple(int, int)
        :param coord: Coordinates in the tables (1-indexed).

        :type  cell: benker.cell.Cell
        :param cell: Cell to insert.
        """
        self._grid[coord] = cell
        # get the cell with its new coordinates:
        cell = self._grid[coord]
        self._refresh_views(cell)

    def __delitem__(self, coord):
        """
        Delete a cell from the table at the given coordinates.

        :type  coord: Coord or tuple(int, int)
        :param coord: Coordinates in the tables (1-indexed).
        """
        del self._grid[coord]
        self._refresh_views()

    def __getitem__(self, coord):
        """
        Get a cell at the given coordinates.

        :type  coord: Coord or tuple(int, int)
        :param coord: Coordinates in the tables (1-indexed).

        :rtype: benker.cell.Cell
        :return: The cell.
        """
        return self._grid[coord]

    def __len__(self):
        return len(self._grid)

    def __iter__(self):
        return iter(self._grid)

    def merge(self, start, end, content_appender=None):
        self._grid.merge(start, end, content_appender=content_appender)
        self._refresh_views()

    def expand(self, coord, width=0, height=0, content_appender=None):
        self._grid.expand(coord, width=width, height=height, content_appender=content_appender)
        self._refresh_views()

    def fill_missing(self, bounding_box, content, styles=None, nature=None):
        """
        Fill the missing cells in the table.

        This method is useful when some rows has missing cells (holes).

        :type  bounding_box: Box
        :param bounding_box:
            The bounding box delimiting the cells/rows to fill if missing.

        :param content:
            User-defined cell content. It can be of any type: ``None``,
            :class:`str`, :class:`int`, :class:`float`, a container (:class:`list`),
            a XML element, etc. The same content can be shared by several cells, it's
            your own responsibility to handle the copy (or deep copy) of the *content*
            reference when needed.

        :type  styles: typing.Dict[str, str]
        :param styles:
            User-defined cell styles: a dictionary of key-value pairs.
            This values are useful to store some HTML-like styles (border-style,
            border-width, border-color, vertical-align, text-align, etc.).
            Of course, we are not tied to the HTML-like styles, you can use your own
            list of styles.

        :type nature: str
        :ivar nature: a way to distinguish the body cells, from the header and the footer.
            The default value is ``None``, but you can use "body", "header", "footer" or whatever
            is suitable for your needs.
            If set to ``None``, the cell nature is inherited from the row nature.

        .. versionadded:: 0.5.0
        """
        for y in range(bounding_box.min.y, bounding_box.max.y + 1):
            for x in range(bounding_box.min.x, bounding_box.max.x + 1):
                if (x, y) not in self:
                    cell = Cell(content, styles=styles, nature=nature, x=x, y=y)
                    self[(x, y)] = cell



# coding: utf-8
"""
CALS to Formex 4 converter
==========================

.. versionadded:: 0.5.0
"""


from benker.builders.formex import FormexBuilder
from benker.converters.base_converter import BaseConverter
from benker.parsers.cals import CalsParser


class Cals2FormexConverter(BaseConverter):
    """
    CALS to Formex 4 converter
    """

    parser_cls = CalsParser
    builder_cls = FormexBuilder


def convert_cals2formex(src_xml, dst_xml, **options):
    """
    Convert CALS 4 tables to Formex tables.

    :param str src_xml:
        Source path of the XML file to convert.

    :param str dst_xml:
        Destination path of the XML file to produce.

    :keyword options:
        Dictionary of parsing/building options.

        **Common parsing options:**

        ``encoding`` (default: "utf-8"):
            XML encoding of the destination file.

        **CALS parser options:**

        ``cals_ns`` (default ``None``):
            Namespace to use for CALS elements and attributes parsing.
            Set ``None`` (or "") if you don't use namespace in your XML.

        ``width_unit`` (default: "mm"):
            Unit to use for table/column widths.
            Possible values are: 'cm', 'dm', 'ft', 'in', 'm', 'mm', 'pc', 'pt', 'px'.

        **Formex 4 builder options:**

        ``use_cals`` (default: ``False``):
            Generate additional CALS-like elements and attributes
            to simplify the layout of Formex document in typesetting systems.

        ``cals_ns`` (default: "https://lib.benker.com/schemas/cals.xsd"):
            Namespace to use for CALS-like elements and attributes (requires: ``use_cals``).
            Set ``None`` (or "") if you don't want to use namespace.

        ``cals_prefix`` (default: "cals"):
            Namespace prefix to use for CALS-like elements and attributes (requires: ``use_cals``).

        ``width_unit`` (default: "mm"):
            Unit to use for table/column widths (requires: ``use_cals``).
            Possible values are: 'cm', 'dm', 'ft', 'in', 'm', 'mm', 'pc', 'pt', 'px'.
    """
    converter = Cals2FormexConverter()
    converter.convert_file(src_xml, dst_xml, **options)



# coding: utf-8
from lxml import etree

from benker.common.lxml_qname import QName

LXML_VERSION = tuple(map(int, etree.__version__.split('.')))


def test_qname__tag():
    qname = etree.QName("tag")
    assert qname.localname == "tag"
    assert qname.namespace is None
    assert qname.text == "tag"


def test_qname__qname():
    qname = etree.QName("{http://somewhere}tag")
    assert qname.localname == "tag"
    assert qname.namespace == "http://somewhere"
    assert qname.text == "{http://somewhere}tag"


def test_qname__element():
    qname = etree.QName(etree.XML('<tag xmlns="http://somewhere"/>'))
    assert qname.localname == "tag"
    assert qname.namespace == "http://somewhere"
    assert qname.text == "{http://somewhere}tag"


def test_qname__uri_tag():
    qname = etree.QName("http://somewhere", "tag")
    assert qname.localname == "tag"
    assert qname.namespace == "http://somewhere"
    assert qname.text == "{http://somewhere}tag"


def test_qname__none_tag():
    qname = etree.QName(None, "tag")
    if LXML_VERSION >= (4, 0):
        assert qname.localname == "tag"
        assert qname.namespace is None
        assert qname.text == "tag"
    else:
        assert qname.localname == "tag"
        assert qname.namespace == "None"
        assert qname.text == "{None}tag"


def test_my_qname__none_tag():
    qname = QName(None, "tag")
    assert qname.localname == "tag"
    assert qname.namespace is None
    assert qname.text == "tag"



# coding: utf-8
import pytest
from lxml import etree

from benker.common.lxml_iterwalk import iterwalk


@pytest.mark.parametrize(
    'events, tag, expected',
    [
        ((), None, []),
        (('start',), None, [('start', 'root'), ('start', 'a')]),
        (('start',), '*', [('start', 'root'), ('start', 'a')]),
        (('start',), 'root', [('start', 'root')]),
        (('start',), 'other', []),
        (('end',), None, [('end', 'a'), ('end', 'root')]),
        (('end',), '*', [('end', 'a'), ('end', 'root')]),
        (('end',), 'root', [('end', 'root')]),
        (('end',), 'other', []),
        (('start', 'end'), None, [('start', 'root'), ('start', 'a'), ('end', 'a'), ('end', 'root')]),
        (('start', 'end'), '*', [('start', 'root'), ('start', 'a'), ('end', 'a'), ('end', 'root')]),
        (('start', 'end'), 'root', [('start', 'root'), ('end', 'root')]),
        (('start', 'end'), 'other', []),
    ],
)
def test_iterwalk_root(events, tag, expected):
    root = etree.XML("<root><a/></root>")
    context = iterwalk(root, events=events, tag=tag)
    actual = [(event, node.tag) for event, node in context]
    assert expected == actual



# coding: utf-8
"""
Cell Size
=========

A :class:`~benker.size.Size` object is used to represent the *width* and *height* of a :class:`~benker.cell.Cell`.
The *width* is the number of spanned columns and the *height* is the number of spanned rows.
The default cell size is (1, 1).

This module defines the :class:`~benker.size.Size` tuple and give some classic use cases.
"""
import collections

SizeTuple = collections.namedtuple('SizeTuple', ['width', 'height'])


class Size(SizeTuple):
    """
    Size of a cell: *width* is the number of columns and *height* is the number of row.

    Usage:

    >>> from benker.size import Size

    >>> size = Size(2, 1)
    >>> size
    Size(width=2, height=1)
    >>> size.width
    2
    >>> size.height
    1
    >>> str(size)
    '(2 x 1)'

    You can use the "+" or "-" operators to increase or decrease the size:

    >>> Size(2, 1) + Size(3, 3)
    Size(width=5, height=4)
    >>> Size(5, 4) - Size(3, 3)
    Size(width=2, height=1)

    You can expand the *width* and *height* to a given factor using the "*" operator:

    >>> Size(2, 1) * 2
    Size(width=4, height=2)

    You can have negative or positive sizes using the unary operators "-" and "+":

    >>> +Size(3, 2)
    Size(width=3, height=2)
    >>> -Size(3, 2)
    Size(width=-3, height=-2)

    .. note::

       A :class:`~benker.cell.Cell` object cannot have a negative or nul sizes,
       but you can need this values for calculation, for instance when you want to
       reduce the cell size.
    """
    __slots__ = ()

    def __str__(self):
        return "({width} x {height})".format(width=self.width, height=self.height)

    def __repr__(self):
        return super(Size, self).__repr__().replace('SizeTuple', 'Size')

    def __add__(self, size):
        size_type = type(size)
        if size_type is Size:
            return Size(self.width + size.width, self.height + size.height)
        elif size_type is tuple:
            return Size(self.width + size[0], self.height + size[1])
        elif size_type is int:
            return Size(self.width + size, self.height + size)
        else:
            raise TypeError(repr(size_type))

    def __sub__(self, size):
        size_type = type(size)
        if size_type is Size:
            return Size(self.width - size.width, self.height - size.height)
        elif size_type is tuple:
            return Size(self.width - size[0], self.height - size[1])
        elif size_type is int:
            return Size(self.width - size, self.height - size)
        else:
            raise TypeError(repr(size_type))

    def __mul__(self, size):
        size_type = type(size)
        if size_type is int:
            return Size(self.width * size, self.height * size)
        else:
            raise TypeError(repr(size_type))

    def __neg__(self):
        return Size(-self.width, -self.height)

    def __pos__(self):
        return Size(self.width, self.height)

    @classmethod
    def from_value(cls, value):
        """
        Convert a value of type :class:`tuple` to a :class:`~benker.size.Size` tuple.

        :param value: tuple of two integers or :class:`~benker.size.Size` tuple.

        :return: Newly created object.

        :raises TypeError:
            if the value is not a tuple of integers nor a :class:`~benker.size.Size` tuple.
        """
        value_type = type(value)
        if value_type is cls:
            return value
        elif value_type is tuple and tuple(map(type, value)) == (int, int):
            return cls(*value)
        raise TypeError(repr(value_type))



# coding: utf-8
"""
Base Converter
==============

Bas class of all converters.
"""
from benker.builders.base_builder import BaseBuilder
from benker.parsers.base_parser import BaseParser


class BaseConverter(object):
    """
    Bas class of all converters.
    """

    parser_cls = BaseParser
    builder_cls = BaseBuilder

    def convert_file(self, src_xml, dst_xml, **options):
        """
        Convert the tables from one format to another.

        :param str src_xml:
            Source path of the XML file to convert.

        :param str dst_xml:
            Destination path of the XML file to produce.

        :keyword options:
            Dictionary of parsing/building options.

            **Common parsing options:**

            ``encoding`` (default: "utf-8"):
                XML encoding of the destination file.
        """
        builder = self.builder_cls(**options)
        parser = self.parser_cls(builder, **options)
        parser.parse_file(src_xml, dst_xml)



import collections
import io
import os
import re

from setuptools import find_packages
from setuptools import setup

install_requires = [
    "lxml",
]

test_require = [
    'PyTest             ; python_version >= "3.6"',
    'PyTest < 5         ; python_version < "3"',
    'PyTest-Cov         ; python_version >= "3.6"',
    'PyTest-Cov < 2.6   ; python_version < "3"',
    "xmldiff",
]

dev_require = [
    "Tox",
    "isort",
    "check-manifest",
    "sphinx < 2",
    'requests[security] ; python_version < "3"',
]

extras_require = {"test": test_require, "dev": dev_require}


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(r":[a-z]+:`~?(.*?)`", u"``\\1``", fd.read())


setup(
    name="Benker",
    version="0.5.3",

    author="Laurent LAPORTE",
    author_email="laurent.laporte.pro@gmail.com",

    description="Easily convert your CALS, HTML, Formex4, Office Open XML (docx) tables from one format to another.",
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    exclude_package_data={"benker": []},
    zip_safe=True,

    url="https://github.com/laurent-laporte-pro/benker",
    project_urls=collections.OrderedDict(
        (
            ("Documentation", "https://benker.readthedocs.io"),
            ("Source Code", "https://github.com/laurent-laporte-pro/benker"),
            ("Issue tracker", "https://github.com/laurent-laporte-pro/benker/issues"),
        )
    ),

    license="MIT",
    platforms=["posix", "nt"],
    keywords="Office, Word, Excel, PowerPoint, docx, xlsx, pptx, CALS, HTML, Formex, table, converter, conversion",

    install_requires=install_requires,
    extras_require=extras_require,

    # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: Markup",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Text Processing :: Markup :: XML",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Topic :: Printing",
        "Operating System :: MacOS",
        "Operating System :: Microsoft",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Other",
        "Operating System :: Unix",
    ],
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
)



# coding: utf-8
"""
Namespace
=========
"""
import collections

from benker.common.lxml_qname import QName


class Namespace(collections.namedtuple("Namespace", "prefix, uri")):
    """
    A namespace is defined by a *prefix* and an *uri*.

    .. versionadded:: 0.5.0
    """

    def get_qname(self, name):
        """ get the qualified name """
        return QName(self.uri, name)

    def get_name(self, name):
        """ get the prefixed name """
        fmt = "{prefix}:{name}" if self.prefix else "{name}"
        return fmt.format(prefix=self.prefix, name=name)



# coding: utf-8
"""
Office Open XML to Formex 4 converter
=====================================

.. versionchanged:: 0.5.0
   Refactoring (rename "Formex4" to "Formex"):

   - the class ``Ooxml2Formex4Converter`` is renamed ``Ooxml2FormexConverter``,
   - the function ``convert_ooxml2formex4`` is renamed ``convert_ooxml2formex``,
"""
from benker.builders.formex import FormexBuilder
from benker.converters.base_converter import BaseConverter
from benker.parsers.ooxml import OoxmlParser


class Ooxml2FormexConverter(BaseConverter):
    """
    Office Open XML to Formex 4 converter
    """

    parser_cls = OoxmlParser
    builder_cls = FormexBuilder


def convert_ooxml2formex(src_xml, dst_xml, **options):
    """
    Convert Office Open XML (OOXML) tables to Formex 4 tables.

    :param str src_xml:
        Source path of the XML file to convert.

        This must be an XML file, for instance, if you want to convert a Word
        document (``.docx``), you first need to unzip the ``.docx`` file, and
        then, run this function on the file ``word/document.xml``.
        You can also use the *styles_path* option to parse and use the table
        styles defined in the file ``word/styles.xml``.

    :param str dst_xml:
        Destination path of the XML file to produce.

    :keyword options:
        Dictionary of parsing/building options.

        **Common parsing options:**

        ``encoding`` (default: "utf-8"):
            XML encoding of the destination file.

        **OOXML parser options:**

        ``styles_path`` (default: ``None``):
            Path to the stylesheet to use to resole table styles.
            In an uncompressed ``.docx`` tree structure, the stylesheet path
            is ``word/styles.xml``.

        **Formex 4 builder options:**

        ``detect_titles`` (default: ``False``):
            If this option is enable, a title will be generated if the first row
            contains an unique cell with centered text.

        ``use_cals`` (default: ``False``):
            Generate additional CALS-like elements and attributes
            to simplify the layout of Formex document in typesetting systems.

        ``cals_ns`` (default: "https://lib.benker.com/schemas/cals.xsd"):
            Namespace to use for CALS-like elements and attributes (requires: ``use_cals``).
            Set ``None`` (or "") if you don't want to use namespace.

        ``cals_prefix`` (default: "cals"):
            Namespace prefix to use for CALS-like elements and attributes (requires: ``use_cals``).

        ``width_unit`` (default: "mm"):
            Unit to use for column widths (requires: ``use_cals``).
            Possible values are: 'cm', 'dm', 'ft', 'in', 'm', 'mm', 'pc', 'pt', 'px'.
    """
    converter = Ooxml2FormexConverter()
    converter.convert_file(src_xml, dst_xml, **options)



# coding: utf-8
"""
OOXML Shading
=============

Table/Cell shading and Table Shading Exception
"""
from lxml import etree

from benker.parsers.ooxml.namespaces import value_of
from benker.parsers.ooxml.types import StHexColor


class Shd(object):
    """
    Table/Cell shading and Table Shading Exception

    Example: ``<w:shd w:val="clear" w:color="auto" w:fill="E6E6E6"/>``

    See: http://www.datypic.com/sc/ooxml/e-w_shd-3.html
    See: http://www.datypic.com/sc/ooxml/e-w_shd-4.html
    See: http://www.datypic.com/sc/ooxml/e-w_shd-5.html
    """

    def __init__(self, w_shd):
        # type: (etree._Element) -> None

        #: Shading Pattern
        self.w_val = value_of(w_shd, "@w:val")  # required

        #: Shading Pattern Color
        self.w_color = value_of(w_shd, "@w:color")

        #: Shading Background Color
        self.w_fill = value_of(w_shd, "@w:fill")

        #: Shading Pattern Theme Color
        self.w_themeColor = value_of(w_shd, "@w:themeColor")

        #: Shading Pattern Theme Color Tint
        self.w_themeTint = value_of(w_shd, "@w:themeTint")

        #: Shading Pattern Theme Color Shade
        self.w_themeShade = value_of(w_shd, "@w:themeShade")

        #: Shading Background Theme Color
        self.w_themeFill = value_of(w_shd, "@w:themeFill")

        #: Shading Background Theme Color Tint
        self.w_themeFillTint = value_of(w_shd, "@w:themeFillTint")

        #: Shading Background Theme Color Shade
        self.w_themeFillShade = value_of(w_shd, "@w:themeFillShade")

    @property
    def styles(self):
        styles = {"background-color": StHexColor(self.w_fill).style}
        return {k: v for k, v in styles.items() if v is not None}



# coding: utf-8
"""
Formex 4 Builder
================

This module can construct a Formex 4 table from
an instance of type :class:`~benker.table.Table`.

:abbr:`Formex (Formalized Exchange of Electronic Publications)` describes the format for the exchange
of data between the Publication Office and its contractors. In particular, it defines the logical
markup for documents which are published in the different series of the Official Journal
of the European Union.

This builder allow you to convert Word document tables
into Formex 4 tables using the Formex 4 schema (formex-05.59-20170418.xd).

Specifications and examples:

- The Formex 4 documentation and schema is available online in the Publication Office:
  `Formex Version 4 <https://op.europa.eu/en/web/eu-vocabularies/formex>`_.

- An example of Formex 4 table is available in the Schema documentation:
  `TBL <https://publications.europa.eu/documents/3938058/5910419/formex_manual_on_screen_version.html/#TBL>`_

.. versionchanged:: 0.5.0
   Refactoring (rename "Formex4" to "Formex"):

   - the class ``Formex4Builder`` is renamed ``FormexBuilder``,
"""
import collections
import re

from lxml import etree

from benker.builders.base_builder import BaseBuilder
from benker.builders.cals import get_colsep_attr
from benker.builders.cals import get_frame_attr
from benker.builders.cals import get_rowsep_attr
from benker.common.lxml_iterwalk import iterwalk
from benker.common.lxml_qname import QName
from benker.schemas import CALS_NS
from benker.schemas import CALS_PREFIX
from benker.units import convert_value
from benker.units import parse_width

# noinspection PyProtectedMember
#: ElementTree Type
ElementTreeType = etree._ElementTree

# noinspection PyProtectedMember
#: Element Type
ElementType = etree._Element

# noinspection PyProtectedMember
#: ProcessingInstructionType is an alias for type hints
ProcessingInstructionType = etree._ProcessingInstruction  # pylint: disable=C0103,W0212

text_type = type(u"")


def revision_mark(name, attrs):
    target = etree.tounicode(etree.Element(name, attrib=attrs))
    target = target.replace("<{0} ".format(name), "").replace("/>", "")
    rev_pi = etree.ProcessingInstruction(name, target)
    return rev_pi


RowInfo = collections.namedtuple("RowInfo", "tag, type, level")


def guess_row_info(rowstyle):
    if rowstyle is None:
        return RowInfo("ROW", None, 0)
    mo = re.match(
        r"""^
        (ROW | TI\.BLK | STI\.BLK)
        (?: - (ALIAS|HEADER|NORMAL|NOTCOL|TOTAL) )?
        (?: - level(\d+) )?
        $""",
        rowstyle,
        flags=re.VERBOSE,
    )
    if mo:
        info_tag = mo.group(1)
        info_type = mo.group(2)
        info_level = int(mo.group(3) or "0")
        return RowInfo(info_tag, info_type, info_level)
    return RowInfo("ROW", None, 0)


class FormexBuilder(BaseBuilder):
    """
    Formex 4 builder used to convert tables into ``TBL`` elements according to the
    `TBL Schema <https://publications.europa.eu/documents/3938058/5910419/formex_manual_on_screen_version.html/#TBL>`_
    """

    # fmt: off
    def __init__(
        self,
        detect_titles=False,
        use_cals=False,
        cals_ns=CALS_NS,
        cals_prefix=CALS_PREFIX,
        width_unit="mm",
        **options
    ):
        # fmt: on
        """
        Initialize the builder.

        :param detect_titles:
            If this option is enable, a title will be generated if the first row
            contains an unique cell with centered text.

        :param bool use_cals:
            Generate additional CALS-like elements and attributes
            to simplify the layout of Formex document in typesetting systems.

        :param str or None cals_ns:
            Namespace to use for CALS-like elements and attributes (requires: ``use_cals``).
            Set ``None`` (or "") if you don't want to use namespace.

        :param str or None cals_prefix:
            Namespace prefix to use for CALS-like elements and attributes (requires: ``use_cals``).

        :param str width_unit:
            Unit to use for column widths (requires: ``use_cals``).
            Possible values are: 'cm', 'dm', 'ft', 'in', 'm', 'mm', 'pc', 'pt', 'px'.

        :keyword options: Extra conversion options.
            See :meth:`~benker.converters.base_converter.BaseConverter.convert_file`
            to have a list of all possible options.

        .. versionchanged:: 0.5.1
           Add the *detect_titles* option.
        """
        # Internal state of the table used during building
        self._table = None
        self._table_colsep = u"0"  # for cals
        self._table_rowsep = u"0"  # for cals

        # NO.SEQ counter
        self._no_seq = 0

        # options
        self.detect_titles = detect_titles
        self.use_cals = use_cals
        self.cals_ns = cals_ns or None
        self.cals_prefix = cals_prefix or None
        self.width_unit = width_unit

        super(FormexBuilder, self).__init__(**options)

    @property
    def ns_map(self):
        if self.use_cals and self.cals_ns:
            return {self.cals_prefix: self.cals_ns}
        return {}

    def get_cals_qname(self, name):
        return QName(self.cals_ns, name)

    def get_formex_qname(self, name):
        # note: in the future, we will use a Formex namespace
        return QName(None, name)

    def generate_table_tree(self, table):
        """
        Build the XML table from the Table instance.

        :type  table: benker.table.Table
        :param table: Table

        :return: Table tree
        """
        table_elem = self.build_tbl(table)
        return table_elem

    def setup_table(self, table):
        self._table = table
        self._table_colsep = u"0"
        self._table_rowsep = u"0"
        return self._table  # mainly for unit tests

    def build_tbl(self, table):
        """
        Build the Formex 4 ``<TBL>`` element.

        Formex 4 attributes:

        -   ``@NO.SEQ`` This mandatory attribute provides a sequence number to the table.
            This number represents the order in which the table appears in the document.

        -   ``@CLASS`` The CLASS attribute is mandatory and is used to specify the type of data
            contained in the table. The allowed values are:

            *   GEN: if the table contains general data (default value),
            *   SCHEDULE: if it is a schedule,
            *   RECAP: if it is a synoptic table.

            These two last values are only used for documents related to the general budget.

        -   ``@COLS`` This mandatory attribute provides the actual number of columns of the table.

        -   ``@PAGE.SIZE`` The PAGE.SIZE attribute takes one of these values:

            *   DOUBLE.LANDSCAPE: table on two A4 pages forming an A3 landscape page,
            *   DOUBLE.PORTRAIT: table on two A4 pages forming an A3 portrait page,
            *   SINGLE.LANDSCAPE: table on a single A4 page in landscape,
            *   SINGLE.PORTRAIT: table on a single A4 page in portrait (default).

        :type  table: benker.table.Table
        :param table: Table

        :return: The newly-created ``<TBL>`` element.

        .. versionchanged:: 0.5.0
           Add support for CALS-like elements and attributes.
           Add support for ``bgcolor`` (Table background color).
        """
        self.setup_table(table)
        table_styles = table.styles
        self._no_seq += 1
        attrs = {
            "NO.SEQ": u"{:04d}".format(self._no_seq),
            # 'CLASS': u"GEN",  # default value
            "COLS": str(len(table.cols)),
            # 'PAGE.SIZE': u"SINGLE.PORTRAIT",  # default value
        }
        if "x-sect-orient" in table_styles:
            # size = (width x height) in 'pt'
            size = table_styles.get("x-sect-size", (595, 842))
            # C4 format 22.9cm x 32.4cm is a little bigger than A4
            if (size[0] <= 649 and size[1] <= 918) or (size[0] <= 918 and size[1] <= 649):
                orient_page_sizes = {"landscape": "SINGLE.LANDSCAPE"}
            else:
                orient_page_sizes = {"landscape": "DOUBLE.LANDSCAPE", "portrait": "DOUBLE.PORTRAIT"}
            orient = table_styles["x-sect-orient"]
            if orient in orient_page_sizes:
                attrs["PAGE.SIZE"] = orient_page_sizes[orient]

        table_elem = etree.Element(u"TBL", attrib=attrs, nsmap=self.ns_map)

        self.build_corpus(table_elem, table)
        return table_elem

    def build_corpus(self, tbl_elem, table):
        """
        Build the Formex 4 ``<CORPUS>`` element.

        :type  tbl_elem: ElementType
        :param tbl_elem: Parent element: ``<TBL>``.

        :type  table: benker.table.Table
        :param table: Table

        .. versionchanged:: 0.5.1
            If this option *detect_titles* is enable, a title will be generated
            if the first row contains an unique cell with centered text.

        .. versionchanged:: 0.5.1
           Add support for the ``@width`` CALS-like attribute (table width).
        """
        table_styles = table.styles
        attrs = {}  # no attribute
        rows = list(table.rows)

        if self.detect_titles:
            # Does the first row/cell contains a centered title?
            first_cell = table[(1, 1)]
            align = first_cell.styles.get("align")
            if first_cell.width == table.bounding_box.width and align == "center":
                # yes, we can generate the title
                self.build_title(tbl_elem, rows.pop(0))

        # support for CALS-like elements and attributes
        if self.use_cals:
            cals = self.get_cals_qname
            attrs[cals("frame")] = get_frame_attr(table_styles)
            self._table_colsep = attrs[cals("colsep")] = get_colsep_attr(table_styles) or "0"
            self._table_rowsep = attrs[cals("rowsep")] = get_rowsep_attr(table_styles) or "0"
            if table.nature is not None:
                attrs[cals("tabstyle")] = table.nature
            if "x-sect-orient" in table_styles:
                attrs[cals("orient")] = {"landscape": "land", "portrait": "port"}[table_styles["x-sect-orient"]]
            if "x-sect-cols" in table_styles:
                attrs[cals("pgwide")] = "1" if table_styles["x-sect-cols"] == "1" else "0"
            if "background-color" in table_styles:
                attrs[cals("bgcolor")] = table_styles["background-color"]
            if "width" in table_styles:
                width, unit = parse_width(table_styles["width"])
                value = convert_value(width, unit, self.width_unit)
                attrs[cals("width")] = u"{value:0.2f}{unit}".format(value=value, unit=self.width_unit)

        corpus_elem = etree.SubElement(tbl_elem, u"CORPUS", attrib=attrs)

        # support for CALS-like elements and attributes
        if self.use_cals:
            for col in table.cols:
                self.build_colspec(corpus_elem, col)

        for row in rows:
            self.build_row(corpus_elem, row)

    def build_title(self, tbl_elem, row):
        """
        Build the table title using the ``<TITLE>`` element.

        For instance:

        .. code-block:: xml

           <TITLE>
             <TI>
               <P>Table IV</P>
             </TI>
           </TITLE>

        :type  tbl_elem: ElementType
        :param tbl_elem: Parent element: ``<TBL>``.

        :type  row: benker.table.RowView
        :param row: The row which contains the title.

        .. versionchanged:: 0.4.4
           Modification of the Formex4 builder to better deal with empty cells (management of ``<IE/>`` tags).
        """
        title_elem = etree.SubElement(tbl_elem, u"TITLE")
        for cell in row.owned_cells:
            # When a cell is empty, we need to insert the ``<IE/>`` tag.
            is_empty_cell = cell.styles.get("x-cell-empty", "false") == "true"
            # We can also have an empty content if a short row is completed by empty cells.
            if is_empty_cell or cell.content is None or cell.content == "":
                # assert cell.content in {None, "", []}
                ti_elem = etree.SubElement(title_elem, u"TI")
                etree.SubElement(ti_elem, u"IE")
            else:
                if isinstance(cell.content, type(u"")):
                    # mainly useful for unit test
                    ti_elem = etree.SubElement(title_elem, u"TI")
                    p_elem = etree.SubElement(ti_elem, u"P")
                    p_elem.text = cell.content
                else:
                    paragraphs = list(cell.content)
                    ti_elem = etree.SubElement(title_elem, u"TI")
                    ti_elem.append(paragraphs[0])
                    sti_elem = etree.SubElement(title_elem, u"STI")
                    sti_elem.extend(paragraphs[1:])

    def build_colspec(self, group_elem, col):
        """
        Build the CALS ``<colspec>`` element (only is *use_cals* is ``True``).

        CALS attributes:

        -   ``@colnum`` is the column number.

        -   ``@colname`` is the column name. Its format is "c{col_pos}".

        -   ``@colwidth`` width of the column (with its unit).
            The unit is defined by the *width_unit* options.

        -   ``@align`` horizontal alignment of table entry content.
            Possible values are: "left", "right", "center", "justify" ("char" is not supported).

        -   ``@colsep`` column separators (vertical ruling).
            Possible values are "0" or "1".

        -   ``@colsep`` row separators (horizontal ruling).
            Possible values are "0" or "1".

        .. note::

           The ``@colnum`` attribute (number of column) is not generated
           because this value is usually implied, and can be deduce
           from the ``@colname`` attribute.

        :type  group_elem: ElementType
        :param group_elem: Parent element: ``<tgroup>``.

        :type  col: benker.table.ColView
        :param col: Columns

        .. versionchanged:: 0.5.0
           Add support for CALS-like elements and attributes.

        .. versionchanged:: 0.5.1
           Add support for CALS-like attributes: ``@colnum``, ``@align``, ``@colsep``, and ``@rowsep``.
        """
        col_styles = col.styles
        cals = self.get_cals_qname

        # -- @cals:colnum
        # -- @cals:colname
        attrs = {cals(u"colnum"): u"{0}".format(col.col_pos), cals(u"colname"): u"c{0}".format(col.col_pos)}

        # -- @cals:colwidth
        if "width" in col_styles:
            width, unit = parse_width(col_styles["width"])
            value = convert_value(width, unit, self.width_unit)
            attrs[cals("colwidth")] = u"{value:0.2f}{unit}".format(value=value, unit=self.width_unit)

        # -- @cals:align
        align = col_styles.get("align")
        align_map = {"left": "left", "right": "right", "center": "center", "justify": "justify"}
        if align in align_map:
            attrs[cals("align")] = align_map[align]

        cell_colsep = get_colsep_attr(col_styles, "border-right")
        if cell_colsep and cell_colsep != self._table_colsep:
            attrs[cals("colsep")] = cell_colsep

        cell_rowsep = get_rowsep_attr(col_styles, "border-bottom")
        if cell_rowsep and cell_rowsep != self._table_rowsep:
            attrs[cals("rowsep")] = cell_rowsep

        etree.SubElement(group_elem, cals(u"colspec"), attrib=attrs, nsmap=self.ns_map)

    def build_row(self, corpus_elem, row):
        """
        Build the Formex 4 ``<ROW>`` element.

        Formex 4 attributes:

        -   ``@TYPE`` The TYPE attribute indicates the specific role of the row in the table.
            The allowed values are:

            *   ALIAS: if the row contains aliases. Such references may be used when the table
                is included on several pages of a publication. The references are associated
                to column headers on the first page and are repeated on subsequent pages.
            *   HEADER: if the row contains cells which may be considered as a column header.
                This generally occurs for the first row of a table.
            *   NORMAL: if most of the cells of the row contain 'simple' or 'normal' data.
                This is the default value.
            *   NOTCOL: if the cells of the row contain units of measure relating to subsequent rows.
            *   TOTAL: if the row contains data which could be considered as 'totals'.

            Note that this TYPE attribute is also provided for the cells (CELL), which could
            be used to override the value defined for the row. On the other hand, 'NORMAL'
            is the default value, so it is necessary to specify the TYPE attribute value
            in each cell of a row which has a specific type in order to avoid the default
            overriding (see the first row of the example below).

        :type  corpus_elem: ElementType
        :param corpus_elem: Parent element: ``<CORPUS>``.

        :type  row: benker.table.RowView
        :param row: The row.

        .. versionchanged:: 0.5.0
           Add support for CALS-like elements and attributes.

        .. versionchanged:: 0.5.1
           The ``@cals:valign`` attribute is built from the "vertical-align" style.
        """
        row_styles = row.styles
        attrs = {}
        nature_types = {"header": u"HEADER", "footer": u"__GR.NOTES__"}
        if row.nature in nature_types:
            attrs["TYPE"] = nature_types[row.nature]

        # support for CALS-like elements and attributes
        if self.use_cals:
            cals = self.get_cals_qname
            if "vertical-align" in row_styles:
                # same values as CSS/Properties/vertical-align
                # fmt: off
                attrs[cals('valign')] = {
                    'top': 'top',
                    'middle': 'middle',
                    'bottom': 'bottom',
                    'baseline': 'bottom',
                }[row_styles['vertical-align']]
                # fmt: on
            row_rowsep = get_rowsep_attr(row_styles, "border-bottom")
            if row_rowsep and row_rowsep != self._table_rowsep:
                attrs[cals("rowsep")] = row_rowsep
            if "rowstyle" in row_styles:
                attrs[cals("rowstyle")] = row_styles["rowstyle"]
            if "background-color" in row_styles:
                attrs[cals("bgcolor")] = row_styles["background-color"]

        if "x-ins" in row_styles:
            # <?change-start change-id="ct140446841083680" type="row:insertion"
            #   creator="Anita BARREL" date="2017-11-15T11:46:00"?>
            rev_attrs = collections.OrderedDict({'type': 'row:insertion'})
            if 'x-ins-id' in row_styles:
                rev_attrs['change-id'] = "ct{0}".format(row_styles['x-ins-id'])
            if 'x-ins-author' in row_styles:
                rev_attrs['creator'] = row_styles['x-ins-author']
            if 'x-ins-date' in row_styles:
                rev_attrs['date'] = row_styles['x-ins-date']
            rev_pi = revision_mark("change-start", rev_attrs)
            corpus_elem.append(rev_pi)

        row_elem = etree.SubElement(corpus_elem, u"ROW", attrib=attrs)

        if "x-ins" in row_styles:
            # <?change-end change-id="ct139821811327752" type="row:insertion"?>
            rev_attrs = collections.OrderedDict({'type': 'row:insertion'})
            if 'x-ins-id' in row_styles:
                rev_attrs['change-id'] = "ct{0}".format(row_styles['x-ins-id'])
            rev_pi = revision_mark('change-end', rev_attrs)
            corpus_elem.append(rev_pi)

        for cell in row.owned_cells:
            self.build_cell(row_elem, cell, row)

    # noinspection PyMethodMayBeStatic
    def build_cell(self, row_elem, cell, row):
        """
        Build the Formex 4 ``<CELL>`` element.

        Formex 4 attributes:

        -   ``@COL`` The mandatory COL attribute is used to specify in which column the cell is located.

        -   ``@COLSPAN`` When a cell in a row 'A' must be linked to a group of cells in the same row,
            the first CELL element of this group has to provide the COLSPAN attribute. The value of
            the COLSPAN attribute is the number of cells in the group. The COL attribute of the first
            cell indicates the number of the first column in the group.

            The use of the COLSPAN attribute is only allowed to relate the value of a cell in several
            columns within the same row. Its value must be at least equal to '2'.

        -   ``@ROWSPAN`` When a cell in column 'A' is linked to a cell in row 'B' located just below
            row 'A', the CELL element of this single cell must provide the ROWSPAN attribute.
            The value of the ROWSPAN attribute is equal to the number of cells in the group.
            The CELL element relating to the single cell must be placed within the first ROW element
            in the group. The ROW elements corresponding to the other rows in the group may not
            contain any CELL elements for the column containing the single cell 'A'.

            The use of the ROWSPAN attribute is only authorised to relate the value of a cell in several
            rows. Its value must be at least equal to '2'.

        -   ``@ACCH`` If the group of related cells is physically delimited by a horizontal brace,
            this symbol must be marked up using the ACCH attribute.

        -   ``@ACCV`` If the group of related cells is physically delimited by a vertical brace,
            this symbol must be marked up using the ACCV attribute.

        -   ``@TYPE`` The TYPE attribute of the CELL element is used to indicate locally the type
            of contents of the cells. It overrides the value of the TYPE attribute defined for
            the row (ROW) which contains the cell.

        :type  row_elem: ElementType
        :param row_elem: Parent element: ``<ROW>``.

        :type  cell: benker.cell.Cell
        :param cell: The cell.

        :type  row: benker.table.RowView
        :param row: The parent row.

        .. versionchanged:: 0.4.4
           Modification of the Formex4 builder to better deal with empty cells (management of ``<IE/>`` tags).

        .. versionchanged:: 0.5.0
           Add support for CALS-like elements and attributes.
           Add support for ``bgcolor`` (Table background color).

        .. versionchanged:: 0.5.1
           Preserve processing instruction in cell content.

        .. versionchanged:: 0.5.2
           Add support for the ``@cals:cellstyle`` attribute (extension).
           This attribute is required for two-way conversion of Formex tables to CALS and vice versa.
           If the ``CELL/@TYPE`` and the ``ROW/@TYPE`` are different, we add a specific "cellstyle" style.
           This style will keep the ``CELL/@TYPE`` value.
        """
        cell_styles = cell.styles
        attrs = {"COL": str(cell.box.min.x)}
        if cell.nature and cell.nature != row.nature:
            nature_types = {"header": u"HEADER", "body": u"NORMAL", "footer": u"__GR.NOTES__"}
            attrs["TYPE"] = nature_types[cell.nature]
        if cell.width > 1:
            attrs[u"COLSPAN"] = str(cell.width)
        if cell.height > 1:
            attrs[u"ROWSPAN"] = str(cell.height)

        # support for CALS-like elements and attributes
        if self.use_cals:
            cals = self.get_cals_qname
            if cell.box.max.x != self._table.bounding_box.max.x:
                # generate @colsep if the cell isn't in the last column
                cell_colsep = get_colsep_attr(cell_styles, "border-right")
                if cell_colsep and cell_colsep != self._table_colsep:
                    attrs[cals("colsep")] = cell_colsep
            if cell.box.max.y != self._table.bounding_box.max.y:
                # generate @rowsep if the cell isn't in the last row
                cell_rowsep = get_rowsep_attr(cell_styles, "border-bottom")
                if cell_rowsep and cell_rowsep != self._table_rowsep:
                    attrs[cals("rowsep")] = cell_rowsep
            if "vertical-align" in cell_styles:
                # same values as CSS/Properties/vertical-align
                # 'w-both' is an extension of OoxmlParser
                attrs[cals('valign')] = {
                    'top': u'top',
                    'middle': u'middle',
                    'bottom': u'bottom',
                    'baseline': u'bottom',
                    'w-both': u'bottom',
                }[cell_styles['vertical-align']]
            if 'align' in cell_styles:
                # same values as CSS/Properties/text-align
            # fmt: off
                attrs[cals('align')] = {
                    'left': u'left',
                    'center': u'center',
                    'right': u'right',
                    'justify': u'justify',
                }[cell_styles['align']]
            # fmt: on
            if cell.width > 1:
                attrs[cals("namest")] = u"c{0}".format(cell.box.min.x)
                attrs[cals("nameend")] = u"c{0}".format(cell.box.max.x)
            if cell.height > 1:
                attrs[cals("morerows")] = str(cell.height - 1)
            if "background-color" in cell_styles:
                attrs[cals("bgcolor")] = cell_styles["background-color"]
            # -- attribute @cals:cellstyle (extension)
            if "cellstyle" in cell_styles:
                # override the value set by the *nature*
                attrs["TYPE"] = cell_styles["cellstyle"]

        cell_elem = etree.SubElement(row_elem, u"CELL", attrib=attrs)

        # When a cell is empty, we need to insert the ``<IE/>`` tag.
        is_empty_cell = cell.styles.get("x-cell-empty", "false") == "true"
        # We can also have an empty content if a short row is completed by empty cells.
        if is_empty_cell or cell.content is None or cell.content == "":
            # The IE element is used to explicitly indicate
            # that specific structures have an empty content.
            etree.strip_tags(cell_elem, "*")
            etree.SubElement(cell_elem, u"IE")
        else:
            self.append_cell_elements(cell_elem, cell.content)

    def finalize_tree(self, tree):
        """
        Finalize the resulting tree structure:

        - Calculate the ``@NO.SEQ`` values: sequence number of each table;
        - Cleanup the ``TBL`` elements when they are direct children of another ``TBL``;
        - Extract ``GR.NOTES`` from the table footers;
        - Group ``ROW`` elements by ``BLK`` based on the ``@cals:rowstyle`` attribute (CALS extension).

        :type  tree: ElementTreeType
        :param tree: The resulting tree.

        .. versionchanged:: 0.5.1
           Drop superfluous CALS-like attributes at the end of the Formex building.
        """
        fmx_root = tree.getroot()
        self.update_no_seq(fmx_root)
        self.cleanup_tbl_in_tbl(fmx_root)
        self.extract_gr_notes(fmx_root)
        self.insert_blk(fmx_root)
        self.drop_superfluous_attrs(fmx_root)

    # noinspection PyMethodMayBeStatic
    def update_no_seq(self, fmx_root):
        """
        Calculate the ``@NO.SEQ`` values: sequence number of each table.

        :type  fmx_root: ElementType
        :param fmx_root: The result tree which contains the ``TBL`` elements to update.
        """
        context = iterwalk(fmx_root, events=("start",), tag=("TBL",))
        stack = []
        for action, elem in context:  # type: str, ElementType
            elem_level = int(elem.xpath("count(ancestor-or-self::TBL)"))
            curr_level = len(stack)
            if curr_level < elem_level:
                stack.extend([0] * (elem_level - curr_level))
            else:
                stack[:] = stack[:elem_level]
            stack[elem_level - 1] += 1
            no_seq = u".".join(u"{:04d}".format(value) for value in stack)
            elem.attrib["NO.SEQ"] = no_seq

    # noinspection PyMethodMayBeStatic
    def cleanup_tbl_in_tbl(self, fmx_root):
        """
        Cleanup the ``TBL`` elements when they are direct children of another ``TBL``

        :type  fmx_root: ElementType
        :param fmx_root: The result tree which contains the ``TBL`` elements to remove.
        """
        for fmx_tbl in fmx_root.xpath("//TBL"):  # type: ElementType
            fmx_parent = fmx_tbl.getparent()
            if fmx_parent is not None and fmx_parent.tag == "TBL":
                # note that we cannot use etree.strip_tags() because a TBL may contains another TBL.
                index = fmx_parent.index(fmx_tbl)
                fmx_parent[index:index + 1] = fmx_tbl.getchildren()

    # noinspection PyPep8Naming
    def extract_gr_notes(self, fmx_root):
        """
        Extract ``GR.NOTES`` from the table footers.

        This function moves or creates a ``GR.NOTES`` just before the ``CORPUS``.

        :type  fmx_root: ElementType
        :param fmx_root: The result tree with ``GR.NOTES``.

        .. versionchanged:: 0.5.1
           If the ROW contains a GR.NOTES, we move it before the CORPUS, else we create it.
        """
        fmx = self.get_formex_qname
        TBL = fmx("TBL").text
        CORPUS = fmx("CORPUS").text
        ROW = fmx("ROW").text
        CELL = fmx("CELL").text
        GR_NOTES = fmx("GR.NOTES").text
        TYPE = fmx("TYPE").text

        for fmx_row in fmx_root.iter(ROW):
            row_type = fmx_row.attrib.pop(TYPE, None)
            if row_type and row_type == '__GR.NOTES__':
                # Get the CORPUS index: the GR.NOTES will be inserted just before the CORPUS
                fmx_tbl = next(fmx_row.iterancestors(TBL))
                index = fmx_tbl.index(fmx_tbl.find(CORPUS))

                # Find the first CELL of the ROW (there is only one CELL)
                fmx_cell = fmx_row.find(CELL)
                cell_text = fmx_cell.text or ""

                fmx_gr_notes = next(fmx_cell.iterchildren(GR_NOTES), None)
                if fmx_gr_notes is None:
                    # insert a new GR.NOTES
                    fmx_gr_notes = etree.Element(GR_NOTES)
                    fmx_gr_notes.text = cell_text
                    fmx_gr_notes.extend(fmx_cell[:])
                    fmx_tbl.insert(index, fmx_gr_notes)
                else:
                    # moves the GR.NOTES (and the potential processing instructions)
                    fmx_tbl[index:index] = fmx_cell[:]
                    if index == 0:
                        fmx_tbl.text = (fmx_tbl.text or "") + cell_text
                    else:
                        prev_node = fmx_tbl[index-1]
                        prev_node.tail = (prev_node.tail or "") + cell_text

                # Add the CALS-like attributes
                fmx_gr_notes.attrib.update((k, v) for k, v in fmx_row.attrib.items() if k not in fmx_gr_notes.attrib)

                # drop the ROW
                fmx_row.getparent().remove(fmx_row)

    def insert_blk(self, fmx_root):
        """
        Group ``ROW`` elements by ``BLK`` based on the ``@cals:rowstyle`` attribute (CALS extension).

        :type  fmx_root: ElementType
        :param fmx_root: The result tree which contains the ``CORPUS/ROW`` elements.
        """
        cals = self.get_cals_qname
        for fmx_corpus in fmx_root.xpath("//CORPUS"):  # type: ElementType
            stack = [fmx_corpus]
            for fmx_row in fmx_corpus.getchildren():  # type: ElementType
                if isinstance(fmx_row, ProcessingInstructionType):
                    # handle PIs (e.g.: revision marks)
                    fmx_top = stack[-1]
                    fmx_top.append(fmx_row)
                    continue
                rowstyle = fmx_row.get(cals("rowstyle"))
                row_info = guess_row_info(rowstyle)
                while len(stack) < row_info.level + 1:
                    fmx_top = stack[-1]
                    fmx_blk = etree.SubElement(fmx_top, "BLK")
                    stack.append(fmx_blk)
                while len(stack) > row_info.level + 1:
                    stack.pop()
                fmx_top = stack[-1]
                if row_info.tag == "ROW":
                    if row_info.type is not None:
                        fmx_row.set("TYPE", row_info.type)
                        for fmx_cell in fmx_row.xpath("CELL"):
                            fmx_cell.set("TYPE", row_info.type)
                    fmx_top.append(fmx_row)
                elif row_info.tag in {"TI.BLK", "STI.BLK"}:
                    # get the COL.START/COL.END integer values (**required**)
                    # find the first non-empty cell (usually this is the first one).
                    for col_pos, fmx_cell in enumerate(fmx_row.xpath("CELL"), 1):
                        if fmx_cell.xpath("self::CELL[string() != '' or count(IE) != 0]"):
                            name_start = fmx_cell.attrib.get(cals("namest"))
                            name_end = fmx_cell.attrib.get(cals("nameend"))
                            if name_start and name_end:
                                col_start = int(re.findall(r"\d+", name_start)[0])
                                col_end = int(re.findall(r"\d+", name_end)[0])
                            else:
                                col_start = col_end = col_pos
                            attrib = {"COL.START": text_type(col_start), "COL.END": text_type(col_end)}
                            break
                    else:
                        # unlikely to go here
                        fmx_cell = fmx_row.xpath("CELL")[0]
                        attrib = {"COL.START": u"1", "COL.END": u"1"}
                    attrib.update({k: v for k, v in fmx_row.attrib.items() if k != "TYPE"})
                    fmx_title = etree.SubElement(fmx_top, row_info.tag, attrib=attrib)
                    fmx_title.text = fmx_cell.text
                    fmx_title.extend(fmx_cell.getchildren())
                    fmx_corpus.remove(fmx_row)
                else:  # pragma: no cover
                    raise NotImplementedError(row_info.tag)

    def drop_superfluous_attrs(self, fmx_root):
        """
        Drop superfluous CALS-like attributes at the end of the Formex building.

        - ``@cals:namest`` and ``@cals:nameend`` are defined by ``@COLSPAN``
        - ``@cals:morerows`` is defined by ``@ROWSPAN``
        - ``@cals:rowstyle`` is defined by ``ROW/@TYPE``, ``GR.NOTES``, ``TI.BLK`` or ``STI.BLK``.

        :type  fmx_root: ElementType
        :param fmx_root: Root element of the Formex file.

        .. versionadded:: 0.5.1
        """
        cals = self.get_cals_qname
        fmx = self.get_formex_qname
        ROW = fmx("ROW").text
        CELL = fmx("CELL").text
        GR_NOTES = fmx("GR.NOTES").text
        TI_BLK = fmx("TI.BLK").text
        STI_BLK = fmx("STI.BLK").text
        elements = {ROW, CELL, TI_BLK, STI_BLK, GR_NOTES}
        for fmx_corpus in fmx_root.xpath("//TBL"):  # type: ElementType
            context = iterwalk(fmx_corpus, events=("start",), tag=elements)
            for action, elem in context:
                elem.attrib.pop(cals("namest"), None)
                elem.attrib.pop(cals("nameend"), None)
                elem.attrib.pop(cals("morerows"), None)
                elem.attrib.pop(cals("rowstyle"), None)
